from abc import abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

from pyjackson import dumps, loads
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text, UniqueConstraint, Float
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from algolink.core.objects import DatasetType
from algolink.core.objects.artifacts import ArtifactCollection
from algolink.core.objects.core import (Buildable, EvaluationResults, EvaluationSet, Image, Model, Pipeline,
                                       PipelineStep, Project, RuntimeEnvironment, RuntimeInstance, Task)
from algolink.core.objects.dataset_source import DatasetSource
from algolink.core.objects.metric import Metric
from algolink.core.objects.requirements import Requirements

SQL_OBJECT_FIELD = '_sqlalchemy_object'


def json_column():
    return Column(Text)


def safe_loads(payload, as_class):
    return loads(payload, Optional[as_class])


def sqlobject(obj):
    return getattr(obj, SQL_OBJECT_FIELD, None)


def update_attrs(obj, **attrs):
    for name, value in attrs.items():
        setattr(obj, name, value)


T = TypeVar('T')
S = TypeVar('S', bound='Attaching')


class Attaching:
    id = ...
    name = ...

    def attach(self, obj):
        setattr(obj, SQL_OBJECT_FIELD, self)
        return obj

    @classmethod
    def from_obj(cls: Type[S], obj: T, new=False) -> S:
        kwargs = cls.get_kwargs(obj)
        existing = sqlobject(obj)
        if not new and existing is not None:
            update_attrs(existing, **kwargs)
            return existing
        return cls(**kwargs)

    @classmethod
    @abstractmethod
    def get_kwargs(cls, obj: T) -> dict:
        pass  # pragma: no cover

    @abstractmethod
    def to_obj(self) -> T:
        pass  # pragma: no cover


Base = declarative_base()


class SProject(Base, Attaching):
    __tablename__ = 'projects'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    tasks: Iterable['STask'] = relationship("STask", back_populates="project")

    def to_obj(self) -> Project:
        p = Project(self.name, id=self.id, author=self.author, creation_date=self.creation_date)
        for task in self.tasks:
            p._tasks.add(task.to_obj())
        return self.attach(p)

    @classmethod
    def get_kwargs(cls, project: Project) -> dict:
        return dict(id=project.id,
                    name=project.name,
                    author=project.author,
                    creation_date=project.creation_date,
                    tasks=[STask.from_obj(t) for t in project.tasks.values()])


class STask(Base, Attaching):
    __tablename__ = 'tasks'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)

    project = relationship("SProject", back_populates="tasks")
    models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    images: Iterable['SImage'] = relationship("SImage", back_populates='task')

    datasets = Column(Text)
    metrics = Column(Text)
    evaluation_sets = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'project_id', name='tasks_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))


class SModel(Base, Attaching):
    __tablename__ = 'models'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    wrapper = Column(Text)

    artifact = Column(Text)
    requirements = Column(Text)
    description = Column(Text)
    params = Column(Text)
    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="models")

    evaluations = Column(Text)
    __table_args__ = (UniqueConstraint('name', 'task_id', name='models_name_and_ref'),)

    def to_obj(self) -> Model:
        model = Model(name=self.name,
                      wrapper_meta=safe_loads(self.wrapper, dict),
                      author=self.author,
                      creation_date=self.creation_date,
                      artifact=safe_loads(self.artifact, ArtifactCollection),
                      requirements=safe_loads(self.requirements, Requirements),
                      description=self.description,
                      params=safe_loads(self.params, Dict[str, Any]),
                      id=self.id,
                      task_id=self.task_id,
                      evaluations=safe_loads(self.evaluations, Dict[str, EvaluationResults]))
        return self.attach(model)

    @classmethod
    def get_kwargs(cls, model: Model) -> dict:
        return dict(id=model.id,
                    name=model.name,
                    author=model.author,
                    creation_date=model.creation_date,
                    wrapper=dumps(model.wrapper_meta),
                    artifact=dumps(model.artifact),
                    requirements=dumps(model.requirements),
                    description=model.description,
                    params=dumps(model.params),
                    task_id=model.task_id,
                    evaluations=dumps(model.evaluations))


class SPipeline(Base, Attaching):
    __tablename__ = 'pipelines'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    steps = Column(Text)

    input_data = Column(Text)
    output_data = Column(Text)

    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="pipelines")

    evaluations = Column(Text)
    __table_args__ = (UniqueConstraint('name', 'task_id', name='pipelines_name_and_ref'),)

    def to_obj(self) -> Pipeline:
        pipeline = Pipeline(name=self.name,
                            steps=safe_loads(self.steps, List[PipelineStep]),
                            input_data=safe_loads(self.input_data, DatasetType),
                            output_data=safe_loads(self.output_data, DatasetType),
                            author=self.author,
                            creation_date=self.creation_date,
                            id=self.id,
                            task_id=self.task_id,
                            evaluations=safe_loads(self.evaluations, EvaluationResults))
        return self.attach(pipeline)

    @classmethod
    def get_kwargs(cls, pipeline: Pipeline) -> dict:
        return dict(id=pipeline.id,
                    name=pipeline.name,
                    author=pipeline.author,
                    creation_date=pipeline.creation_date,
                    steps=dumps(pipeline.steps),
                    input_data=dumps(pipeline.input_data),
                    output_data=dumps(pipeline.output_data),
                    task_id=pipeline.task_id,
                    evaluations=dumps(pipeline.evaluations))


class SImage(Base, Attaching):
    __tablename__ = 'images'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="images")

    environment_id = Column(Integer, ForeignKey('environments.id'), nullable=False)

    params = Column(Text)
    source = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'task_id', name='image_name_and_ref'),)

    def to_obj(self) -> Image:
        image = Image(name=self.name,
                      author=self.author,
                      creation_date=self.creation_date,
                      id=self.id,
                      task_id=self.task_id,
                      params=safe_loads(self.params, Image.Params),
                      source=safe_loads(self.source, Buildable),
                      environment_id=self.environment_id)
        return self.attach(image)

    @classmethod
    def get_kwargs(cls, image: Image) -> dict:
        return dict(id=image.id,
                    name=image.name,
                    author=image.author,
                    creation_date=image.creation_date,
                    task_id=image.task_id,
                    params=dumps(image.params),
                    source=dumps(image.source),
                    environment_id=image.environment_id)


class SRuntimeEnvironment(Base, Attaching):
    __tablename__ = 'environments'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    params = Column(Text)

    def to_obj(self) -> RuntimeEnvironment:
        environment = RuntimeEnvironment(
            name=self.name,
            author=self.author,
            creation_date=self.creation_date,
            id=self.id,
            params=safe_loads(self.params, RuntimeEnvironment.Params))
        return self.attach(environment)

    @classmethod
    def get_kwargs(cls, environment: RuntimeEnvironment) -> dict:
        return dict(id=environment.id,
                    name=environment.name,
                    author=environment.author,
                    creation_date=environment.creation_date,
                    params=dumps(environment.params))


class SRuntimeInstance(Base, Attaching):
    __tablename__ = 'instances'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    image_id = Column(Integer, ForeignKey('images.id'), nullable=False)
    environment_id = Column(Integer, ForeignKey('environments.id'), nullable=False)

    params = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'image_id', 'environment_id', name='instance_name_and_ref'),)

    def to_obj(self) -> RuntimeInstance:
        instance = RuntimeInstance(
            name=self.name,
            author=self.author,
            creation_date=self.creation_date,
            id=self.id,
            image_id=self.image_id,
            environment_id=self.environment_id,
            params=safe_loads(self.params, RuntimeInstance.Params))
        return self.attach(instance)

    @classmethod
    def get_kwargs(cls, instance: RuntimeInstance) -> dict:
        return dict(id=instance.id,
                    name=instance.name,
                    author=instance.author,
                    creation_date=instance.creation_date,
                    image_id=instance.image_id,
                    environment_id=instance.environment_id,
                    params=dumps(instance.params))

class TExperiment(Base, Attaching):
    __tablename__ = 'experiments'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    project_remark = Column(String, unique=False, nullable=False)
    nick_name = Column(String, unique=False, nullable=False)
    del_flag = Column(Integer, default = 0)
    creation_date = Column(DateTime, unique=False, nullable=False)

    submodels: Iterable['TSubmodel'] = relationship("TSubmodel", back_populates="experiment")

    def to_obj(self) -> Project:
        p = Project(self.name, id=self.id, author=self.author, creation_date=self.creation_date)
        for task in self.tasks:
            p._tasks.add(task.to_obj())
        return self.attach(p)

    @classmethod
    def get_kwargs(cls, project: Project) -> dict:
        return dict(id=project.id,
                    name=project.name,
                    author=project.author,
                    creation_date=project.creation_date,
                    tasks=[STask.from_obj(t) for t in project.tasks.values()])

class TSubmodel(Base, Attaching):
    __tablename__ = 'submodels'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    sub_model_remark = Column(String, unique=False, nullable=False)
    nick_name = Column(String, unique=False)
    sub_model_sequence = Column(Integer, nullable=False)
    del_flag = Column(Integer, default = 0)

    creation_date = Column(DateTime, unique=False, nullable=False)
    experiment_id = Column(Integer, ForeignKey('experiments.id'), nullable=False)

    experiment = relationship("TExperiment", back_populates="submodels")
    model_metrics: Iterable['TModelmetric'] = relationship("TModelmetric", back_populates="submodel")
    model_params: Iterable['TModelparam'] = relationship("TModelparam", back_populates="submodel")
    best_result: Iterable['TBestresult'] = relationship("TBestresult", back_populates="submodel")
    #models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    #pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    #images: Iterable['SImage'] = relationship("SImage", back_populates='task')

    #datasets = Column(Text)
    #metrics = Column(Text)
    #evaluation_sets = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'experiment_id', name='submodel_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))

class TModelmetric(Base, Attaching):
    __tablename__ = 'model_metrics'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    metric_type = Column(String, unique=False, nullable=False)
    metric_value = Column(Float, nullable=False)
    epoch = Column(Integer, nullable=False)

    creation_date = Column(DateTime, unique=False, nullable=False)
    sub_model_id = Column(Integer, ForeignKey('submodels.id'), nullable=False)

    submodel = relationship("TSubmodel", back_populates="model_metrics")
    #models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    #pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    #images: Iterable['SImage'] = relationship("SImage", back_populates='task')

    #datasets = Column(Text)
    #metrics = Column(Text)
    #evaluation_sets = Column(Text)

    #__table_args__ = (UniqueConstraint('name', 'sub_model_id', name='metric_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))

class TModelparam(Base, Attaching):
    __tablename__ = 'model_params'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    param_type = Column(String, unique=False, nullable=False)
    param_value = Column(Float, nullable=False)

    creation_date = Column(DateTime, unique=False, nullable=False)
    sub_model_id = Column(Integer, ForeignKey('submodels.id'), nullable=False)

    submodel = relationship("TSubmodel", back_populates="model_params")

    #models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    #pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    #images: Iterable['SImage'] = relationship("SImage", back_populates='task')

    #datasets = Column(Text)
    #metrics = Column(Text)
    #evaluation_sets = Column(Text)

    #__table_args__ = (UniqueConstraint('name', 'sub_model_id', name='metric_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))

class TBestresult(Base, Attaching):
    __tablename__ = 'best_result'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    best_value = Column(Float, nullable=False)
    best_epoch = Column(Integer, nullable=False)

    creation_date = Column(DateTime, unique=False, nullable=False)
    sub_model_id = Column(Integer, ForeignKey('submodels.id'), nullable=False)

    submodel = relationship("TSubmodel", back_populates="best_result")
    #models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    #pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    #images: Iterable['SImage'] = relationship("SImage", back_populates='task')

    #datasets = Column(Text)
    #metrics = Column(Text)
    #evaluation_sets = Column(Text)

    #__table_args__ = (UniqueConstraint('name', 'sub_model_id', name='metric_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))



import logging
from typing import Dict, Type

from ..utils.classproperty import classproperty
from ..utils.importing import import_string

logger = logging.getLogger(__name__)


def registering_type(type_name):
    """
    Helper for base classes which maintains registry of all their subclasses

    :param type_name: name for base class to use
    :return: class with subclasses registry built in
    """

    class RegType:
        impls: Dict[str, Type] = {}

        def __init_subclass__(cls, **kwargs):
            RegType.impls[cls.classpath] = cls
            super(RegType, cls).__init_subclass__(**kwargs)

        @staticmethod
        def get(name):
            import_string(name)
            impl = RegType.impls.get(name)
            if impl is None:
                raise ValueError('{} class {} not found'.format(type_name, name))
            return impl

        @classproperty
        def classpath(cls):
            return '{}.{}'.format(cls.__module__, cls.__name__)

    return RegType



from algolink.core.objects.core import Buildable
from algolink.utils.importing import import_string


class BuildableWithServer(Buildable):
    def __init__(self, server_type: str):
        self.server_type = server_type

    @property
    def server(self):
        return import_string(self.server_type)()



import importlib
import sys
from types import ModuleType
from typing import Dict, List, Union

from ..config import Core
from ..utils.classproperty import classproperty
from ..utils.importing import import_module, module_importable, module_imported
from ..utils.log import logger


class Extension:
    """
    Extension descriptor

    :param module: main extension module
    :param reqs: list of extension dependencies
    :param force: if True, disable lazy loading for this extension
    :param validator: boolean predicate which should evaluate to True for this extension to be loaded
    """

    def __init__(self, module, reqs: List[str], force=True, validator=None):
        self.force = force
        self.reqs = reqs
        self.module = module
        self.validator = validator

    def __str__(self):
        return f'<Extension {self.module}>'

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.module == other.module

    def __hash__(self):
        return hash(self.module)


class ExtensionDict(dict):
    """
    :class:`_Extension` container
    """

    def __init__(self, *extensions: Extension):
        super().__init__()
        for e in extensions:
            self[e.module] = e


def __tensorflow_major_version():
    import tensorflow as tf
    return tf.__version__.split('.')[0]


is_tf_v1, is_tf_v2 = lambda: __tensorflow_major_version() == '1', lambda: __tensorflow_major_version() == '2'


class ExtensionLoader:
    """
    Class that tracks and loads extensions.

    """
    builtin_extensions: Dict[str, Extension] = ExtensionDict(
        Extension('algolink.ext.numpy', ['numpy'], False),
        Extension('algolink.ext.pandas', ['pandas'], False),
        Extension('algolink.ext.sklearn', ['sklearn'], False),
        Extension('algolink.ext.tensorflow', ['tensorflow'], False, is_tf_v1),
        Extension('algolink.ext.tensorflow_v2', ['tensorflow'], False, is_tf_v2),
        Extension('algolink.ext.torch', ['torch'], False),
        Extension('algolink.ext.catboost', ['catboost'], False),
        Extension('algolink.ext.aiohttp', ['aiohttp', 'aiohttp_swagger']),
        Extension('algolink.ext.flask', ['flask', 'flasgger'], False),
        Extension('algolink.ext.sqlalchemy', ['sqlalchemy']),
        Extension('algolink.ext.s3', ['boto3']),
        Extension('algolink.ext.imageio', ['imageio']),
        Extension('algolink.ext.lightgbm', ['lightgbm'], False),
        Extension('algolink.ext.xgboost', ['xgboost'], False),
        Extension('algolink.ext.docker', ['docker'], False)
    )

    _loaded_extensions: Dict[Extension, ModuleType] = {}

    @classproperty
    def loaded_extensions(cls) -> Dict[Extension, ModuleType]:
        """
        :return: List of loaded extensions
        """
        return cls._loaded_extensions

    @classmethod
    def _setup_import_hook(cls, extensions: List[Extension]):
        """
        Add import hook to sys.meta_path that will load extensions when their dependencies are imported

        :param extensions: list of :class:`.Extension`
        """
        if len(extensions) == 0:
            return

        existing = [h for h in sys.meta_path if isinstance(h, _ImportLoadExtInterceptor)]
        if len(existing) > 0:
            hook = existing[0]
            hook.module_to_extension.update({req: e for e in extensions for req in e.reqs})
        else:
            hook = _ImportLoadExtInterceptor(
                module_to_extension={req: e for e in extensions for req in e.reqs}
            )
            sys.meta_path.insert(0, hook)

    @classmethod
    def load_all(cls, try_lazy=True):
        """
        Load all (builtin and additional) extensions

        :param try_lazy: if `False`, use force load for all builtin extensions
        """
        for_hook = []
        for ext in cls.builtin_extensions.values():
            if not try_lazy or hasattr(sys, 'frozen') or ext.force:
                if all(module_importable(r) for r in ext.reqs):
                    cls.load(ext)
            else:
                if all(module_imported(r) for r in ext.reqs):
                    cls.load(ext)
                else:
                    for_hook.append(ext)

        cls._setup_import_hook(for_hook)

        for mod in Core.ADDITIONAL_EXTENSIONS:
            cls.load(mod)

    @classmethod
    def load(cls, extension: Union[str, Extension]):
        """
        Load single extension

        :param extension: str of :class:`.Extension` instance to load
        """
        if isinstance(extension, str):
            extension = Extension(extension, [], force=True)
        if extension not in cls._loaded_extensions and not module_imported(extension.module) and \
                (extension.validator is None or extension.validator()):
            logger.debug('Importing extension module %s', extension.module)
            cls._loaded_extensions[extension] = import_module(extension.module)


class _ImportLoadExtRegisterer(importlib.abc.PathEntryFinder):
    """A hook that registers all modules that are being imported"""

    def __init__(self):
        self.imported = []

    def find_module(self, fullname, path=None):
        self.imported.append(fullname)
        return None


class _ImportLoadExtInterceptor(importlib.abc.Loader, importlib.abc.PathEntryFinder):
    """
    Import hook implementation to load extensions on dependency import

    :param module_to_extension: dict requirement -> :class:`.Extension`
    """

    def __init__(self, module_to_extension: Dict[str, Extension]):
        self.module_to_extension = module_to_extension

    def find_module(self, fullname, path=None):
        # hijack importing machinery
        return self

    def load_module(self, fullname):
        # change this hook to registering hook
        reg = _ImportLoadExtRegisterer()
        sys.meta_path = [reg] + [x for x in sys.meta_path if x is not self]
        try:
            # fallback to ordinary importing
            module = importlib.import_module(fullname)
        finally:
            # put this hook back
            sys.meta_path = [self] + [x for x in sys.meta_path if x is not reg]

        # check all that was imported and import all extensions that are ready
        for imported in reg.imported:
            if not module_imported(imported):
                continue
            extension = self.module_to_extension.get(imported)
            if extension is None:
                continue

            if all(module_imported(m) for m in extension.reqs):
                ExtensionLoader.load(extension)

        return module


def load_extensions(*exts: str):
    """
    Load extensions

    :param exts: list of extension main modules
    """
    for ext in exts:
        ExtensionLoader.load(ext)



from .dataset import FeedDictDatasetType, FeedDictHook
from .model import TFTensorHook, TFTensorModelWrapper

__all__ = ['FeedDictHook', 'FeedDictDatasetType', 'TFTensorHook', 'TFTensorModelWrapper']



from abc import abstractmethod
from typing import List

from algolink.core.analyzer import Hook
from ..objects.requirements import AnyRequirements, Requirement, Requirements, resolve_requirements


class RequirementAnalyzer:
    """
    Analyzer for RequirementHook hooks
    """
    hooks: List['RequirementHook'] = []

    @classmethod
    def analyze(cls, obj: AnyRequirements) -> Requirements:
        """
        Run RequirementHook hooks to analyze obj

        :param obj: objects to analyze
        :return: Instance of Requirements
        """
        obj = resolve_requirements(obj)
        to_process = list(obj.requirements)
        result = []
        for hook in cls.hooks:
            while len(to_process) > 0:
                req = to_process.pop()
                if hook.can_process(req):
                    result += hook.process(req).requirements
                else:
                    result.append(req)
            to_process = result
            result = []

        return Requirements(to_process)


class RequirementHook(Hook):
    _analyzer = RequirementAnalyzer

    def must_process(self, obj: Requirement) -> bool:
        return self.can_process(obj)

    @abstractmethod
    def can_process(self, obj: Requirement) -> bool:
        pass

    @abstractmethod
    def process(self, obj: Requirement, **kwargs) -> Requirements:
        pass



from .dataset import TFTensorDatasetType, TFTensorHook
from .model import TFKerasModelHook, TFKerasModelWrapper

__all__ = ['TFTensorDatasetType', 'TFTensorHook', 'TFKerasModelHook', 'TFKerasModelWrapper']



from algolink.core.analyzer.requirement import RequirementHook
from algolink.core.objects.requirements import InstallableRequirement, Requirement, Requirements, UnixPackageRequirement


class LightGBMRequirementHook(RequirementHook):
    def can_process(self, obj: Requirement) -> bool:
        return isinstance(obj, InstallableRequirement) and obj.package.lower() == 'lightgbm'

    def process(self, obj: Requirement, **kwargs) -> Requirements:
        return Requirements([obj, UnixPackageRequirement('libgomp1')])



import os
import warnings
from functools import reduce
from typing import Optional, Union

from pyjackson import dumps
from pyjackson.decorators import cached_property

from algolink.build.provider import PythonProvider
from algolink.build.provider.utils import BuildableWithServer
from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.buildable import BuildableHook
from algolink.core.objects import ArtifactCollection, Model, Pipeline, Requirements, Task
from algolink.core.objects.artifacts import Blobs, CompositeArtifactCollection, LocalFileBlob, _RelativePathWrapper
from algolink.core.objects.core import _with_meta
from algolink.runtime.interface.pipeline import MODEL_BIN_PATH, PIPELINE_META_PATH, PipelineLoader, PipelineMeta
from algolink.runtime.server import Server
from algolink.utils.module import get_object_requirements

LOADER_PATH = 'loader'
SERVER_PATH = 'server'


def read(path):
    with open(path) as f:
        return f.read()


class PipelineProvider(PythonProvider):
    """Provider to build service from Pipeline object

    :param pipeline: Pipeline instance to build from
    :param server: Server instance to build with
    :param debug: Whether to run image in debug mode
    """

    def __init__(self, pipeline: Pipeline, server: Server, debug: bool = False):
        super().__init__(server, PipelineLoader(), debug)
        self.pipeline = pipeline

    @cached_property
    def _requirements(self) -> Requirements:
        """Union of pipeline, server and loader requirements"""
        model_reqs = reduce(Requirements.__add__,
                            [m.requirements for m in self.pipeline.models.values()],
                            Requirements())
        return (model_reqs +
                get_object_requirements(self.server) +
                get_object_requirements(self.loader))

    def get_requirements(self) -> Requirements:
        """Returns union of model, server and loader requirements"""
        return self._requirements

    def _get_sources(self):
        """Returns sources of custom modules from requirements"""
        sources = {}
        for cr in self.get_requirements().custom:
            sources.update(cr.to_sources_dict())

        # add __init__.py for all dirs that doesnt have it already
        packages = set(os.path.join(os.path.dirname(p), '__init__.py') for p in sources if os.path.dirname(p) != '')
        sources.update({
            p: '' for p in packages if p not in sources
        })
        return sources

    def get_sources(self):
        """Returns model metadata file and sources of custom modules from requirements"""
        meta = PipelineMeta(self.pipeline, {
            k: v.without_artifacts() for k, v in self.pipeline.models.items()
        })
        return {
            PIPELINE_META_PATH: dumps(meta),
            **self._get_sources(),
            **{os.path.basename(f): read(f) for f in self.server.additional_sources}
        }

    def get_artifacts(self) -> ArtifactCollection:
        """Return model binaries"""

        artifacts = CompositeArtifactCollection(
            [_RelativePathWrapper(m.artifact_any, os.path.join(MODEL_BIN_PATH, m.name)) for m in
             self.pipeline.models.values()])
        if len(self.server.additional_binaries) > 0:
            artifacts = CompositeArtifactCollection([
                artifacts,
                Blobs({os.path.basename(f): LocalFileBlob(f) for f in self.server.additional_binaries})
            ])
        return artifacts

    def get_python_version(self):
        """
        :return: version of python that produced model
        """
        versions = {model.params.get(Model.PYTHON_VERSION) for model in self.pipeline.models.values()}
        versions.discard(None)
        if len(versions) == 0:
            return super(PipelineProvider, self).get_python_version()
        elif len(versions) > 1:
            warnings.warn(f'Inconsistent python version for pipeline models: {versions}')
        return min(versions)  # in backward compatibility we trust


class PipelineBuildable(BuildableWithServer):
    def __init__(self, pipeline_id: Union[int, Pipeline], server_type: str, debug: bool = False):
        super().__init__(server_type)
        self.debug = debug
        if isinstance(pipeline_id, int):
            self.pipeline_id = pipeline_id
            self.pipeline_cache = None
        else:
            self.pipeline_cache = pipeline_id
            self.pipeline_id = pipeline_id.id
            self.bind_meta_repo(self.pipeline_cache._meta)

    @property
    def task(self) -> Optional[Task]:
        return self.pipeline.task

    @property
    @_with_meta(False)
    def pipeline(self):
        if self.pipeline_cache is None:
            self._check_meta(False)
            self.pipeline_cache = self._meta.get_pipeline_by_id(self.pipeline_id)
        return self.pipeline_cache

    def get_provider(self) -> PipelineProvider:
        return PipelineProvider(self.pipeline, self.server, self.debug)


class BuildableModelHook(BuildableHook, TypeHookMixin):
    valid_types = [Pipeline]

    def process(self, obj, **kwargs):
        server = kwargs.get('server')  # TODO ???
        debug = kwargs.get('debug', False)
        return PipelineBuildable(obj, server.type, debug)



import builtins
from abc import abstractmethod
from typing import Dict, List, Sized

from pyjackson import deserialize, serialize
from pyjackson.core import ArgList, Field
from pyjackson.decorators import type_field
from pyjackson.errors import DeserializationError, SerializationError

from .base import AlgoLinkParams
from .requirements import InstallableRequirement, Requirements
from .typing import SizedTypedListType, TypeWithSpec


@type_field('type')
class DatasetType(AlgoLinkParams, TypeWithSpec):
    """
    Base class for dataset type metadata.
    Children of this class must be both pyjackson-serializable and be a pyjackson serializer for it's dataset type
    """
    type = None

    @staticmethod
    def _check_type(obj, exp_type, exc_type):
        if not isinstance(obj, exp_type):
            raise exc_type(f'given dataset is of type: {type(obj)}, expected: {exp_type}')

    @property
    @abstractmethod
    def requirements(self) -> Requirements:
        """"""  # TODO docs

    @abstractmethod
    def get_writer(self):
        """"""  # TODO docs


class LibDatasetTypeMixin(DatasetType):
    """
    :class:`.DatasetType` mixin which provides requirements list consisting of
    PIP packages represented by module objects in `libraries` field.
    """
    libraries = None

    @property
    def requirements(self) -> Requirements:
        return Requirements([InstallableRequirement.from_module(lib) for lib in self.libraries])


PRIMITIVES = {int, str, bool, complex, float}


class PrimitiveDatasetType(DatasetType):
    """
    DatasetType for int, str, bool, complex and float types
    """
    type = 'primitive'

    def __init__(self, ptype: str):
        self.ptype = ptype

    @classmethod
    def from_object(cls, obj):
        if type(obj) not in PRIMITIVES:
            raise ValueError('{} type is not primitive')
        return PrimitiveDatasetType(str(type(obj)))

    @property
    def to_type(self):
        return getattr(builtins, self.ptype)

    def get_spec(self) -> ArgList:
        return [Field(None, self.to_type, False)]

    def deserialize(self, obj):
        try:
            return self.to_type(obj)
        except (ValueError, TypeError):
            raise DeserializationError(f'given object: {obj} could not be converted to {self.to_type}')

    def serialize(self, instance):
        self._check_type(instance, self.to_type, SerializationError)
        return instance

    @property
    def requirements(self) -> Requirements:
        return Requirements()

    def get_writer(self):
        from algolink.repository.dataset.artifact import PrimitiveDatasetWriter
        return PrimitiveDatasetWriter()


class ListDatasetType(DatasetType, SizedTypedListType):
    """
    DatasetType for list type
    """
    real_type = None
    type = 'list'

    def __init__(self, dtype: DatasetType, size: int):
        SizedTypedListType.__init__(self, size, dtype)

    def deserialize(self, obj):
        _check_type_and_size(obj, list, self.size, DeserializationError)
        return [deserialize(o, self.dtype) for o in obj]

    def serialize(self, instance: list):
        _check_type_and_size(instance, list, self.size, SerializationError)
        return [serialize(o, self.dtype) for o in instance]

    @property
    def requirements(self) -> Requirements:
        return self.dtype.requirements

    def get_writer(self):
        from algolink.repository.dataset.artifact import PickleWriter
        return PickleWriter()


class _TupleLikeDatasetType(DatasetType):
    """
    DatasetType for tuple-like collections
    """
    real_type = None

    def __init__(self, items: List[DatasetType]):
        self.items = items

    def get_spec(self) -> ArgList:
        return [Field(str(i), t, False) for i, t in enumerate(self.items)]

    def deserialize(self, obj):
        _check_type_and_size(obj, self.actual_type, len(self.items), DeserializationError)
        return self.actual_type(deserialize(o, t) for t, o in zip(self.items, obj))

    def serialize(self, instance: Sized):
        _check_type_and_size(instance, self.actual_type, len(self.items), SerializationError)
        return self.actual_type(serialize(o, t) for t, o in zip(self.items, instance))

    @property
    def requirements(self) -> Requirements:
        return sum([i.requirements for i in self.items], Requirements())

    def get_writer(self):
        from algolink.repository.dataset.artifact import PickleWriter
        return PickleWriter()


def _check_type_and_size(obj, dtype, size, exc_type):
    DatasetType._check_type(obj, dtype, exc_type)
    if size != -1 and len(obj) != size:
        raise exc_type(f'given {dtype.__name__} has len: {len(obj)}, expected: {size}')


class TupleLikeListDatasetType(_TupleLikeDatasetType):
    """
    DatasetType for tuple-like list type
    """
    actual_type = list
    type = 'tuple_like_list'


class TupleDatasetType(_TupleLikeDatasetType):
    """
    DatasetType for tuple type
    """
    actual_type = tuple
    type = 'tuple'


class DictDatasetType(DatasetType):
    """
    DatasetType for dict type
    """
    real_type = None
    type = 'dict'

    def __init__(self, item_types: Dict[str, DatasetType]):
        self.item_types = item_types

    def get_spec(self) -> ArgList:
        return [Field(name, t, False) for name, t in self.item_types.items()]

    def deserialize(self, obj):
        self._check_type_and_keys(obj, DeserializationError)
        return {k: deserialize(v, self.item_types[k]) for k, v in obj.items()}

    def serialize(self, instance: dict):
        self._check_type_and_keys(instance, SerializationError)
        return {
            k: serialize(v, self.item_types[k]) for k, v in instance.items()
        }

    def _check_type_and_keys(self, obj, exc_type):
        self._check_type(obj, dict, exc_type)
        if set(obj.keys()) != set(self.item_types.keys()):
            raise exc_type(f'given dict has keys: {set(obj.keys())}, expected: {set(self.item_types.keys())}')

    @property
    def requirements(self) -> Requirements:
        return sum([i.requirements for i in self.item_types.values()], Requirements())

    def get_writer(self):
        from algolink.repository.dataset.artifact import PickleWriter
        return PickleWriter()


class BytesDatasetType(DatasetType):
    """
    DatasetType for bytes objects
    """
    type = 'bytes'
    real_type = None

    def __init__(self):
        pass

    def get_spec(self) -> ArgList:
        return [Field('file', bytes, False)]

    def deserialize(self, obj) -> object:
        return obj

    def serialize(self, instance: object) -> dict:
        return instance

    @property
    def requirements(self) -> Requirements:
        return Requirements()

    def get_writer(self):
        from algolink.repository.dataset.artifact import PickleWriter
        return PickleWriter()



import contextlib
import os
from typing import Dict

import docker
import docker.errors
import requests
from pyjackson.core import Comparable
from pyjackson.decorators import type_field

from algolink.core.objects import Image, RuntimeEnvironment, RuntimeInstance
from algolink.ext.docker.utils import create_docker_client, image_exists_at_dockerhub
from algolink.utils.log import logger


@type_field('type')
class DockerRegistry(Comparable):
    """Registry for docker images. This is the default implementation that represents registry of the docker daemon"""

    def get_host(self) -> str:
        """Returns registry host or emty string for local"""
        return ''

    def push(self, client: docker.DockerClient, tag: str):
        """Pushes image to registry

        :param client: DockerClient to use
        :param tag: name of the tag to push"""

    def login(self, client: docker.DockerClient):
        """Login to registry

        :param client: DockerClient to use"""

    def uri(self, image: str):
        """Cretate an uri for image in this registry

        :param image: image name"""
        return image

    def image_exists(self, client: docker.DockerClient, image: 'DockerImage'):
        """Check if image exists in this registry

        :param client: DockerClient to use
        :param image: :class:`.DockerImage` to check"""
        try:
            client.images.get(image.uri)
            return True
        except docker.errors.ImageNotFound:
            return False

    def delete_image(self, client: docker.DockerClient, image: 'DockerImage', force: bool = False, **kwargs):
        """Deleta image from this registry

        :param client: DockerClient to use
        :param image: :class:`.DockerImage` to delete
        :param force: force delete
        """
        try:
            client.images.remove(image.uri, force=force, **kwargs)
        except docker.errors.ImageNotFound:
            pass


class DockerIORegistry(DockerRegistry):
    """ The class represents docker.io registry.

    """

    def get_host(self) -> str:
        return 'https://index.docker.io/v1/'

    def push(self, client, tag):
        client.images.push(tag)
        logger.info('Pushed image %s to docker.io', tag)

    def image_exists(self, client, image: 'DockerImage'):
        return image_exists_at_dockerhub(image.uri)

    def delete_image(self, client, image: 'DockerImage', force=False, **kwargs):
        logger.warn('Skipping deleting image %s from docker.io', image.name, force, **kwargs)


class RemoteRegistry(DockerRegistry):
    """DockerRegistry implementation for official Docker Registry (as in https://docs.docker.com/registry/)

    :param host: adress of the registry"""

    def __init__(self, host: str = None):
        self.host = host  # TODO credentials?

    def login(self, client):
        """
        Logs in to Docker registry

        Corresponding credentials should be specified as environment variables per registry:
        e.g., if registry host is "168.32.25.1:5000" then
        "168_32_25_1_5000_USERNAME" and "168_32_25_1_5000_PASSWORD" variables should be specified

        :param client: Docker client instance
        :return: nothing
        """

        host_for_env = self.host.replace('.', '_').replace(':', '_')
        username_var = f'{host_for_env}_username'.upper()
        username = os.getenv(username_var)
        password_var = f'{host_for_env}_password'.upper()
        password = os.getenv(password_var)

        if username and password:
            client.login(registry=self.host, username=username, password=password)
            logger.info('Logged in to remote registry at host %s', self.host)
        else:
            logger.warning('Skipped logging in to remote registry at host %s because no credentials given. ' +
                           'You could specify credentials as %s and %s environment variables.',
                           self.host, username_var, password_var)

    def get_host(self) -> str:
        return self.host

    def push(self, client, tag):
        client.images.push(tag)
        logger.info('Pushed image %s to remote registry at host %s', tag, self.host)

    def uri(self, image: str):
        return f'{self.host}/{image}'

    def _get_digest(self, name, tag):
        r = requests.head(f'http://{self.host}/v2/{name}/manifests/{tag}',
                          headers={'Accept': 'application/vnd.docker.distribution.manifest.v2+json'})
        if r.status_code != 200:
            return
        return r.headers['Docker-Content-Digest']

    def image_exists(self, client, image: 'DockerImage'):
        name = image.fullname
        digest = self._get_digest(name, image.tag)
        if digest is None:
            return False
        r = requests.head(f'http://{self.host}/v2/{name}/manifests/{digest}')
        if r.status_code == 404:
            return False
        elif r.status_code == 200:
            return True
        r.raise_for_status()

    def delete_image(self, client, image: 'DockerImage', force=False, **kwargs):
        name = image.fullname
        digest = self._get_digest(name, image.tag)
        if digest is None:
            return
        requests.delete(f'http://{self.host}/v2/{name}/manifests/{digest}')


@type_field('type')
class DockerDaemon(Comparable):
    """Class that represents docker daemon

    :param host: adress of the docker daemon (empty string for local)"""

    def __init__(self, host: str):  # TODO credentials
        self.host = host

    @contextlib.contextmanager
    def client(self) -> docker.DockerClient:
        """Get DockerClient isntance"""
        with create_docker_client(self.host) as c:
            yield c


class DockerImage(Image.Params):
    """:class:`.Image.Params` implementation for docker images
    full uri for image looks like registry.host/repository/name:tag

    :param name: name of the image
    :param tag: tag of the image
    :param repository: repository of the image
    :param registry: :class:`.DockerRegistry` instance with this image
    :param image_id: docker internal id of this image"""

    def __init__(self, name: str, tag: str = 'latest', repository: str = None, registry: DockerRegistry = None,
                 image_id: str = None):
        self.repository = repository
        self.image_id = image_id
        self.name = name
        self.tag = tag
        self.registry = registry or DockerRegistry()

    @property
    def fullname(self):
        return f'{self.repository}/{self.name}' if self.repository is not None else self.name

    @property
    def uri(self) -> str:
        return self.registry.uri(f'{self.fullname}:{self.tag}')

    def exists(self, client: docker.DockerClient):
        """Checks if this image exists in it's registry"""
        return self.registry.image_exists(client, self)

    def delete(self, client: docker.DockerClient, force=False, **kwargs):
        """Deletes image from registry"""
        self.registry.delete_image(client, self, force, **kwargs)


class DockerContainer(RuntimeInstance.Params):
    """:class:`.RuntimeInstance.Params` implementation for docker containers

    :param name: name of the container
    :param port_mapping: port mapping in this container
    :param params: other parameters for docker run cmd
    :param container_id: internal docker id for this container"""

    def __init__(self, name: str, port_mapping: Dict[int, int] = None, params: Dict[str, object] = None,
                 container_id: str = None):
        self.container_id = container_id
        self.name = name
        self.port_mapping = port_mapping or {}
        self.params = params or {}


class DockerEnv(RuntimeEnvironment.Params):
    """:class:`.RuntimeEnvironment.Params` implementation for docker environment

    :param registry: default registry to push images to
    :param daemon: :class:`.DockerDaemon` instance"""

    def __init__(self, registry: DockerRegistry = None, daemon: DockerDaemon = None):
        self.registry = registry or DockerRegistry()
        self.daemon = daemon or DockerDaemon('')

    def get_runner(self):
        """
        :return: docker runner
        """
        if self.default_runner is None:
            from .runner import DockerRunner
            self.default_runner = DockerRunner()
        return self.default_runner

    def get_builder(self):
        """
        :return: docker builder instance
        """
        if self.default_builder is None:
            from .builder import DockerBuilder
            self.default_builder = DockerBuilder()
        return self.default_builder



from .artifact.base import ArtifactRepository
from .dataset.base import DatasetRepository
from .metadata.base import MetadataRepository

__all__ = ['ArtifactRepository', 'MetadataRepository', 'DatasetRepository']



import datetime
import getpass
import json
import re
import tempfile
import time
import warnings
from abc import abstractmethod
from copy import copy
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from pyjackson import deserialize, serialize
from pyjackson.core import Comparable
from pyjackson.decorators import make_string, type_field

import algolink.repository
from algolink.client.expose import ExposedMethod
from algolink.core import errors
from algolink.core.analyzer.metric import MetricAnalyzer
from algolink.core.analyzer.model import ModelAnalyzer
from algolink.core.analyzer.requirement import RequirementAnalyzer
from .artifacts import ArtifactCollection, CompositeArtifactCollection
from .base import AlgoLinkParams
from .dataset_source import AbstractDataset, Dataset, DatasetSource, InMemoryDatasetSource
from .dataset_type import DatasetType
from .metric import Metric
from .requirements import AnyRequirements, Requirements, resolve_requirements
from .wrapper import ModelWrapper, WrapperArtifactCollection
from algolink.utils.index_dict import IndexDict, IndexDictAccessor
from algolink.utils.log import logger
from algolink.utils.module import get_python_version

AnyDataset = Union[str, AbstractDataset, DatasetSource, Any]
AnyMetric = Union[str, Metric, Any]


def _get_current_user():
    return getpass.getuser()


class ExposedObjectMethod(ExposedMethod):
    """Decorator for AlgoLinkObject methods that will be exposed to AlgoLink class by autogen

    :param name: name of the exposed method
    :param param_name: name of the first parameter
    :param param_type: type hint for the first parameter
    :param param_doc: docstring for the first parameter"""

    def __init__(self, name, param_name: str, param_type: str, param_doc: str = None):
        super().__init__(name)
        self.param_name = param_name
        self.param_type = param_type
        self.param_doc = param_doc

    def get_doc(self):
        doc = self.method.__doc__
        if self.param_doc is not None:
            if ':param' in doc:
                mark = ':param'
            elif ':' in doc:
                mark = ':'
            else:
                mark = None
            param_doc = f':param {self.param_name}: {self.param_doc}\n'
            if mark is not None:
                doc = re.sub(mark, param_doc + '        ' + mark, doc, count=1)
            else:
                doc += '\n\n        ' + param_doc + '        '
        return doc

    def generate_code(self):
        """Generates code for exposed AlgoLink method"""
        declaration = self.get_declaration() \
            .replace(f'{self.name}(self', f'{self.name}(self, {self.param_name}: {self.param_type}')
        fields = self.get_signature().args
        result = f'''{self.param_name}.{self.original_name}({', '.join(f.name for f in fields)})'''
        return f'''    {declaration}
        """"{self.get_doc()}"""
        return {result}'''


class WithMetadataRepository:
    """Intermediat abstract class for objects that can be binded to meta repository"""
    _id = None
    _meta: 'algolink.repository.MetadataRepository' = None
    _nested_fields_meta: List[str] = []

    def bind_meta_repo(self, repo: 'algolink.repository.MetadataRepository'):
        self._meta = repo
        for field in self._nested_fields_meta:
            for obj in getattr(self, field).values():
                obj.bind_meta_repo(repo)
        return self

    def unbind_meta_repo(self):
        for field in self._nested_fields_meta:
            for obj in getattr(self, field).values():
                obj.unbind_meta_repo()
        del self._meta
        self._id = None

    @property
    def has_meta_repo(self):
        return self._meta is not None

    def _check_meta(self, saved=True):
        """Checks if object is binded to meta repo

        :param saved: force object to be also saved (have id)"""
        if (self._id is None and saved) or not self.has_meta_repo:
            raise errors.UnboundObjectError('{} is not bound to meta repository'.format(self))


class WithArtifactRepository:
    """Intermediat abstract class for objects that can be binded to artifact repository"""
    _art: 'algolink.repository.ArtifactRepository' = None
    _nested_fields_art: List[str] = []

    def bind_artifact_repo(self, repo: 'algolink.repository.ArtifactRepository'):
        self._art = repo
        for field in self._nested_fields_art:
            for obj in getattr(self, field).values():
                obj.bind_artifact_repo(repo)
        return self

    def unbind_artifact_repo(self):
        for field in self._nested_fields_art:
            for obj in getattr(self, field).values():
                obj.unbind_artifact_repo()
        del self._art

    @property
    def has_artifact_repo(self):
        return self._art is not None


class WithDatasetRepository:
    """Intermediat abstract class for objects that can be binded to dataset repository"""
    _dataset: 'algolink.repository.DatasetRepository' = None
    _nested_fields_dataset: List[str] = []

    def bind_dataset_repo(self, repo: 'algolink.repository.DatasetRepository'):
        self._dataset = repo
        for field in self._nested_fields_dataset:
            for obj in getattr(self, field).values():
                obj.bind_dataset_repo(repo)
        return self

    def unbind_dataset_repo(self):
        for field in self._nested_fields_dataset:
            for obj in getattr(self, field).values():
                obj.unbind_dataset_repo()
        del self._dataset

    @property
    def has_dataset_repo(self):
        return self._dataset is not None


class AlgoLinkObject(Comparable, WithMetadataRepository, WithArtifactRepository):
    """
    Base class for high level algolink objects.
    These objects can be binded to metadata repository and/or to artifact repository

    :param id: object id
    :param name: object name
    :param author: user that created that object
    :param creation_date: date when this object was created
    """

    def __init__(self, id: int, name: str, author: str = None, creation_date: datetime.datetime = None):
        self._id = id
        self.name = name
        self.author = author or _get_current_user()
        self.creation_date = creation_date or datetime.datetime.utcnow()  # TODO local timezone

    @property
    def id(self) -> int:
        return self._id

    def bind_as(self, other: 'AlgoLinkObject'):
        if other.has_meta_repo:
            self.bind_meta_repo(other._meta)
        if other.has_artifact_repo:
            self.bind_artifact_repo(other._art)
        return self

    @abstractmethod
    def save(self):
        """Saves object state to metadata repository"""

    @abstractmethod
    def has_children(self):
        """Checks if object has existing relationship"""


def _with_meta(saved=True):
    """
    Decorator for methods to check that object is binded to meta repo

    :param saved: method to apply decorator
    """

    def dec(method):
        @wraps(method)
        def inner(self: AlgoLinkObject, *args, **kwargs):
            self._check_meta(saved)
            return method(self, *args, **kwargs)

        return inner

    return dec(saved) if callable(saved) else dec


def _with_artifact(method):
    """
    Decorator for methods to check that object is binded to artifact repo

    :param method: method to apply decorator
    :return: decorated method
    """

    @wraps(method)
    def inner(self: AlgoLinkObject, *args, **kwargs):
        if not self.has_artifact_repo:
            raise errors.UnboundObjectError('{} is not bound to artifact repository'.format(self))
        return method(self, *args, **kwargs)

    return inner


def _with_dataset(method):
    """
    Decorator for methods to check that object is binded to dataset repo

    :param method: method to apply decorator
    :return: decorated method
    """

    @wraps(method)
    def inner(self: WithDatasetRepository, *args, **kwargs):
        if not self.has_dataset_repo:
            raise errors.UnboundObjectError('{} is not bound to dataset repository'.format(self))
        return method(self, *args, **kwargs)

    return inner


@make_string('id', 'name')
class Project(AlgoLinkObject):
    """
    Project is a collection of tasks

    :param id: project id
    :param name: project name
    :param author: user that created that project
    :param creation_date: date when this project was created
    """
    _nested_fields_meta = _nested_fields_art = ['_tasks']

    def __init__(self, name: str, id: int = None, author: str = None, creation_date: datetime.datetime = None):
        super().__init__(id, name, author, creation_date)
        self._tasks: IndexDict[Task] = IndexDict('id', 'name')
        self.tasks: IndexDictAccessor[Task] = IndexDictAccessor(self._tasks)

    @_with_meta
    @ExposedObjectMethod(name='delete_project', param_name='project', param_type='Project',
                         param_doc='project to delete')
    def delete(self, cascade: bool = False):
        """
        Deletes project and(if required) all tasks associated with it from metadata repository

        :param cascade: whether should project be deleted with all associated tasks
        :return: Nothing
        """
        if cascade:
            for task in self._meta.get_tasks(self):  # TODO cacheing?
                self.delete_task(task, cascade=cascade)
                # task.delete(cascade=cascade) # FIXME
        self._meta.delete_project(self)

    @_with_meta
    def add_task(self, task: 'Task'):
        """
        Add task to project and save it to meta repo

        :param task: task to add
        """
        if task.project_id is not None and task.project_id != self.id:
            raise errors.MetadataError('Task is already in project {}. Delete it first'.format(task.project_id))

        task.project_id = self.id
        self._meta.save_task(task)
        self._tasks.add(task)
        return task.bind_as(self)

    @_with_meta
    def add_tasks(self, tasks: List['Task']):
        """
        Add multiple tasks and save them to meta repo

        :param tasks: tasks to add
        """
        for t in tasks:
            self.add_task(t)

    @_with_meta
    def delete_task(self, task: 'Task', cascade: bool = False):
        """
        Remove task from this project and delete it from meta repo

        :param cascade: whether task should be deleted with all nested objects
        :param task: task to delete
        """
        if task.id not in self._tasks:
            raise errors.NonExistingTaskError(task)
        del self._tasks[task.id]
        task.delete(cascade)
        task.project_id = None

    @_with_meta
    def save(self):
        self._meta.save_project(self)

    def has_children(self):
        return len(self.tasks) > 0

    def __repr__(self):
        return """Project '{name}', {td} tasks""".format(name=self.name, td=len(self.tasks))


class EvaluationSet(AlgoLinkParams):
    """Represents a set of objects for evaluation

    :param input_dataset: name of the input dataset
    :param output_dataset: name of the output dataset
    :param metrics: list of metric names"""

    def __init__(self, input_dataset: str, output_dataset: str, metrics: List[str]):
        self.output_dataset = output_dataset
        self.metrics = metrics
        self.input_dataset = input_dataset

    def get(self, task: 'Task', cache=True) -> Tuple[DatasetSource, DatasetSource, Dict[str, Metric]]:
        """Loads actual datasets and metrics from task

        :param task: task to load from
        :param cache: wheter to cache datasets"""
        input = task.datasets[self.input_dataset]
        output = task.datasets[self.output_dataset]
        if cache:
            input = input.cache()
            output = output.cache()
        return input, output, {m: task.metrics[m] for m in self.metrics}

class EvaluationSet(AlgoLinkParams):
    """Represents a set of objects for evaluation

    :param input_dataset: name of the input dataset
    :param output_dataset: name of the output dataset
    :param metrics: list of metric names"""

    def __init__(self, input_dataset: str, output_dataset: str, metrics: List[str]):
        self.output_dataset = output_dataset
        self.metrics = metrics
        self.input_dataset = input_dataset

    def get(self, task: 'Task', cache=True) -> Tuple[DatasetSource, DatasetSource, Dict[str, Metric]]:
        """Loads actual datasets and metrics from task

        :param task: task to load from
        :param cache: wheter to cache datasets"""
        input = task.datasets[self.input_dataset]
        output = task.datasets[self.output_dataset]
        if cache:
            input = input.cache()
            output = output.cache()
        return input, output, {m: task.metrics[m] for m in self.metrics}


@make_string('id', 'name')
class Task(AlgoLinkObject, WithDatasetRepository):
    """
    Task is a collection of models

    :param id: task id
    :param name: task name
    :param project_id: parent project id for this task
    :param author: user that created that task
    :param creation_date: date when this task was created
    """
    _nested_fields_meta = _nested_fields_art = ['_models', '_pipelines', '_images', '_experiments']

    def __init__(self, name: str, id: int = None, project_id: int = None,
                 author: str = None, creation_date: datetime.datetime = None,
                 datasets: Dict[str, DatasetSource] = None,
                 metrics: Dict[str, Metric] = None,
                 evaluation_sets: Dict[str, EvaluationSet] = None):
        super().__init__(id, name, author, creation_date)
        self.evaluation_sets = evaluation_sets or {}
        self.datasets = datasets or {}
        self.metrics = metrics or {}
        self.project_id = project_id
        self._models: IndexDict[Model] = IndexDict('id', 'name')
        self.models: IndexDictAccessor[Model] = IndexDictAccessor(self._models)
        self._pipelines: IndexDict[Pipeline] = IndexDict('id', 'name')
        self.pipelines: IndexDictAccessor[Pipeline] = IndexDictAccessor(self._pipelines)
        self._images: IndexDict[Image] = IndexDict('id', 'name')
        self.images: IndexDictAccessor[Image] = IndexDictAccessor(self._images)
        self._experiments: IndexDict[Experiment] = IndexDict('id', 'name')
        self.experiments: IndexDictAccessor[Experiment] = IndexDictAccessor(self._experiments)

    @property
    @_with_meta
    def project(self):
        p = self._meta.get_project_by_id(self.project_id)
        if p is None:
            raise errors.NonExistingProjectError(self.project_id)
        return p.bind_artifact_repo(self._art)

    @project.setter
    def project(self, project: Project):
        if not isinstance(project, Project):
            raise ValueError('{} is not Project'.format(project))
        self.project_id = project.id
        self.bind_as(project)

    @_with_meta
    @ExposedObjectMethod(name='delete_task', param_name='task', param_type='Task', param_doc='task to delete')
    def delete(self, cascade: bool = False):
        """
        Deletes task from metadata

        :param cascade: whether should task be deleted with all associated objects
        :return: Nothing
        """
        if cascade:
            for model in self._meta.get_models(self):
                self.delete_model(model.bind_artifact_repo(self._art))
            for image in self._meta.get_images(self):
                self.delete_image(image, cascade=cascade)
            for pipeline in self._meta.get_pipelines(self):
                self.delete_pipeline(pipeline)

        for dataset in self.datasets:
            self.delete_dataset(dataset, force=True, save=False)

        self._meta.delete_task(self)

    @_with_meta
    def add_model(self, model: 'Model'):
        """
        Add model to task and save it to meta repo

        :param model: model to add
        """
        if model.task_id is not None and model.task_id != self.id:
            raise errors.MetadataError('Model is already in task {}. Delete it first'.format(model.task_id))

        model.task_id = self.id
        self._meta.save_model(model)
        self._models.add(model)
        return model.bind_as(self)

    @_with_meta
    def add_models(self, models: List['Model']):
        """
        Add multiple models and save them to meta repo

        :param models: models to add
        """
        for m in models:
            self.add_model(m)

    @_with_meta
    def delete_model(self, model: 'Model', force=False):
        """
        Remove model from this task and delete it from meta repo

        :param model: model to delete
        :param force: whether model artifacts' deletion errors should be ignored, default is false
        """
        model_id = model.id
        if model_id not in self._models:
            raise errors.NonExistingModelError(model)

        model.delete(force)
        del self._models[model_id]

    #  ##########API############
    @_with_meta
    @_with_artifact
    def create_and_push_model(self, model_object, input_data, model_name: str = None, **kwargs) -> 'Model':
        """
        Create :class:`Model` instance from model object and push it to repository

        :param model_object: model object to build Model from
        :param input_data: input data sample to determine structure of inputs and outputs for given model
        :param model_name: name for model
        :param kwargs: other :meth:`~Model.create` arguments
        :return: created :class:`Model`
        """
        model = Model.create(model_object, input_data, model_name, **kwargs)
        return self.push_model(model)

    @_with_meta
    @_with_artifact
    def push_model(self, model: 'Model') -> 'Model':
        """
        Push :class:`Model` instance to task repository

        :param model: :class:`Model` to push
        :return: same pushed :class:`Model`
        """
        model.bind_as(self)
        model = model.push(self)
        self._models.add(model)
        return model

    @_with_meta
    def add_pipeline(self, pipeline: 'Pipeline'):
        """
        Add model to task and save it to meta repo

        :param pipeline: pipeline to add
        """
        if pipeline.task_id is not None and pipeline.task_id != self.id:
            raise errors.MetadataError('Pipeline is already in task {}. Delete it first'.format(pipeline.task_id))

        pipeline.task_id = self.id
        self._meta.save_pipeline(pipeline)
        self._pipelines.add(pipeline)

    @_with_meta
    def add_pipelines(self, pipelines: List['Pipeline']):
        """
        Add multiple models and save them to meta repo

        :param pipelines: pipelines to add
        """
        for m in pipelines:
            self.add_pipeline(m)

    @_with_meta
    def delete_pipeline(self, pipeline: 'Pipeline'):
        """
        Remove model from this task and delete it from meta repo

        :param pipeline: pipeline to delete
        """
        pipeline_id = pipeline.id
        if pipeline_id not in self._pipelines:
            raise errors.NonExistingPipelineError(pipeline)

        pipeline.delete()
        del self._pipelines[pipeline_id]
        pipeline.task_id = None

    @_with_meta
    def add_image(self, image: 'Image'):
        """
        Add image for model and save it to meta repo

        :param image: image to add
        """
        if image.task_id is not None and image.task_id != self.id:
            raise errors.MetadataError('Image is already in task {}. Delete it first'.format(image.task_id))

        image.task_id = self.id
        self._meta.save_image(image)
        self._images.add(image)

    @_with_meta
    def add_images(self, images: List['Image']):
        """
        Add multiple images for model and save them to meta repo

        :param images: images to add
        """
        for image in images:
            self.add_image(image)

    @_with_meta
    def delete_image(self, image: 'Image', meta_only: bool = False, cascade: bool = False):
        """
        Remove image from this model and delete it from meta repo

        :param image: image to delete
        :param meta_only: should image be deleted only from metadata
        :param cascade: whether image should be deleted with all instances
        """
        if image.id not in self._images:
            raise errors.NonExistingImageError(image)
        del self._images[image.id]
        image.delete(meta_only=meta_only, cascade=cascade)
        image.task_id = None

    @_with_meta
    def save(self):
        """Saves task to meta repository and pushes unsaved datasets to dataset repository"""
        self.push_datasets()
        self._meta.save_task(self)

    def has_children(self):
        return len(self.models) > 0 or len(self.pipelines) > 0 or len(self.images) > 0

    def _resolve_dataset(self, dataset: AnyDataset, name: str) -> str:
        """Resolves existing dataset or creates a new one

        :param dataset: dataset name, Dataset, DatasetSource or raw data object
        :param name: name for dataset if it is new
        """
        if isinstance(dataset, str):
            if dataset not in self.datasets:
                raise ValueError(f'no dataset named {dataset} in task {self}')  # TODO maybe ohter error?
            return dataset

        self.add_dataset(name, dataset)
        return name

    def _resolve_metric(self, metric: AnyMetric, name: str):
        """Resolves existing metric or creates a new one

        :param metric: metric name, Metric or raw metric object
        :param name: name for metric if it is new
        """
        if isinstance(metric, str):
            if metric not in self.metrics:
                raise ValueError(f'no metric named {metric} in task {self}')  # TODO maybe ohter error?
            return metric
        self.add_metric(name, metric)
        return name

    def add_evaluation(self, name: str,
                       data: AnyDataset,
                       target: AnyDataset,
                       metrics: Union[AnyMetric, List[AnyMetric]]):
        """Adds new evaluation set to this task

        :param name: name of the evaluation set
        :param data: input dataset for evaluation
        :param target: ground truth for input data
        :param metrics: one or more metrics to measure
        """
        if name in self.evaluation_sets:
            raise errors.AlgoLinkError(f'evalset {name} already in task {self}')
        data = self._resolve_dataset(data, f'{name}_input')
        target = self._resolve_dataset(target, f'{name}_output')
        if not isinstance(metrics, list):
            metrics = [metrics]
        metrics = [self._resolve_metric(m, f'{name}_{i}') for i, m in enumerate(metrics)]
        self.evaluation_sets[name] = EvaluationSet(data, target, metrics)

    def delete_evaluation(self, name: str, save: bool = True):
        """Deletes evaluation set from task

        :param name: name of the evaluation to delete
        :param save: also update task metadata in repo
        """
        if save:
            self._check_meta()

        if name not in self.evaluation_sets:
            raise errors.AlgoLinkError(f'cannot delete evalset from task {self}: no evalset with name {name}')
        del self.evaluation_sets[name]
        if save:
            self.save()

    @_with_dataset
    def add_dataset(self, name, dataset: Union[DatasetSource, AbstractDataset, Any]):
        """Adds new dataset to this task

        :param name: name of the dataset
        :param dataset: Dataset, DatasetSource or raw dataset object
        """
        if name in self.datasets:
            raise errors.AlgoLinkError(f'dataset {name} already in task {self}')
        if not isinstance(dataset, DatasetSource):
            if not isinstance(dataset, AbstractDataset):
                dataset = Dataset.from_object(dataset)
            dataset = dataset.to_inmemory_source()
        self.datasets[name] = dataset

    @_with_dataset
    def push_datasets(self):
        """Pushes all unsaved datasets to dataset repository"""
        for name, dataset in list(self.datasets.items()):
            if isinstance(dataset, InMemoryDatasetSource):
                self.datasets[name] = self._dataset.save(f'{self.id}/{name}', dataset.read())

    @_with_dataset
    def delete_dataset(self, name: str, force: bool = False, save: bool = True):
        """Deletes dataset from task with artifacts

        :param name: name of the dataset to delete
        :param force: wheter to check evalsets that use this dataset and remove them or raise error
        :param save: also update task metadata in repo
        """
        if name not in self.datasets:
            raise errors.AlgoLinkError(f'cannot delete dataset from {self}: no dataset with name {name}')
        if save:
            self._check_meta()
        evalsets = {n: es for n, es in self.evaluation_sets.items() if
                    name in (es.input_dataset, es.output_dataset)}
        if len(evalsets) > 0 and not force:
            raise errors.AlgoLinkError(
                f'cannot delete dataset from {self}: it is used by evalsets {list(evalsets.keys())}'
                f'remove them or set force=True')
        try:
            self._dataset.delete(f'{self.id}/{name}')
        except errors.NoSuchDataset:
            pass  # already deleted or was not saved
        del self.datasets[name]
        for es in evalsets.keys():
            self.delete_evaluation(es, save)
        if save:
            self.save()

    def add_metric(self, name, metric: Union[Metric, Any]):
        """Adds metric to this task

        :param name: name of the metric
        :param metric: Metric or raw metric object
        """
        if name in self.metrics:
            raise errors.AlgoLinkError(f'metric {name} already in task {self}')
        if not isinstance(metric, Metric):
            metric = MetricAnalyzer.analyze(metric)
        # TODO checks
        self.metrics[name] = metric

    def delete_metric(self, name: str, force: bool = False, save: bool = True):
        """Deletes metric from task

        :param name: name of the metric to delete
        :param force: wheter to check evalsets that use this metric and remove them or raise error
        :param save: also update task metadata in repo
        """
        if name not in self.metrics:
            raise errors.AlgoLinkError(f'cannot delete metric from {self}: no metric with name {name}')
        if save:
            self._check_meta()
        evalsets = {n: es for n, es in self.evaluation_sets.items() if
                    name in es.metrics}

        if len(evalsets) > 0 and not force:
            raise errors.AlgoLinkError(
                f'cannot delete metric from {self}: it is used by evalsets {list(evalsets.keys())}'
                f'remove them or set force=True')
        del self.metrics[name]
        for es in evalsets.keys():
            self.delete_evaluation(es, save)
        if save:
            self.save()

    def evaluate_all(self, force=False, save_result=True) -> Dict[str, 'EvaluationResult']:
        """Evaluates all viable pairs of evalsets and models/pipelines

        :param force: force reevaluate already evaluated
        :param save_result: save evaluation results to meta"""
        result = {}
        timestamp = time.time()
        for name, evalset in self.evaluation_sets.items():
            input, output, metrics = evalset.get(self, cache=True)
            for model in self._models.values():
                evaluate = model.evaluate(input, output, metrics, evaluation_name=name,
                                          timestamp=timestamp, force=force, save=save_result)
                if isinstance(evaluate, dict):
                    for method, res in evaluate.items():
                        result[f'{model.name}.{method}'] = res
                elif isinstance(evaluate, EvaluationResult):
                    result[model.name] = evaluate
            for pipeline in self._pipelines.values():
                evaluate = pipeline.evaluate(input, output, metrics, evaluation_name=name,
                                             timestamp=timestamp, force=force, save=save_result)
                if evaluate is not None:
                    result[pipeline.name] = evaluate
        return result

    #@_with_meta
    def create_experiment(self, experiment: 'Experiment'):
        """
        Add experiment to task and save it to meta repo

        :param experiment: experiment to add
        """
        if experiment.task_id is not None and experiment.task_id != self.id:
            raise errors.MetadataError('Experiment is already in task {}. Delete it first'.format(experiment.task_id))

        experiment.task_id = self.id
        self._meta.save_experiment(experiment)
        self._experiments.add(experiment)
        return experiment.bind_as(self)

class _InTask(AlgoLinkObject):
    """Intermediate abstract class for object inside task"""

    def __init__(self, id: int, name: str,
                 author: str = None, creation_date: datetime.datetime = None,
                 task_id: int = None):
        super().__init__(id, name, author, creation_date)
        self.task_id = task_id

    @property
    @_with_meta
    def task(self):
        t = self._meta.get_task_by_id(self.task_id)
        if t is None:
            raise errors.NonExistingTaskError(self.task_id)
        return t.bind_artifact_repo(self._art)

    @task.setter
    def task(self, task: Task):
        if not isinstance(task, Task):
            raise ValueError('{} is not Task'.format(task))
        self.task_id = task.id
        self.bind_as(task)


@make_string
class EvaluationResult(AlgoLinkParams):
    """Represents result of evaluation of one evalset on multiple evaluatable objects

    :param scores: mapping 'metric' -> 'score'
    :param timestamp: time of evaluation
    """

    def __init__(self, timestamp: float, scores: Dict[str, float] = None):
        self.timestamp = timestamp
        self.scores = scores or {}


@make_string
class EvaluationResultCollection(AlgoLinkParams):
    """Collection of evaluation results for single evalset

    :param results: list of results"""

    def __init__(self, results: List[EvaluationResult] = None):
        self.results = results or []

    def add(self, result: EvaluationResult):
        if result != self.latest:
            self.results.append(result)

    @property
    def latest(self) -> Optional[EvaluationResult]:
        if len(self.results) == 0:
            return
        return max(self.results, key=lambda r: r.timestamp)


EvaluationResults = Dict[str, EvaluationResultCollection]  # Evaluation results for all evalsets
MultipleResults = Dict[str, EvaluationResult]  # Evaluation results for multiple objects but one evalset


class _WrapperMethodAccessor:
    """Class to access ModelWrapper methods from model

    :param model: model to access
    :param method_name: name of the wrapper method"""

    def __init__(self, model: 'Model', method_name: str):
        if model.wrapper.methods is None or method_name not in model.wrapper.exposed_methods:
            raise AttributeError(f'{model} does not have {method_name} method')
        self.model = model
        self.method_name = method_name

    def __call__(self, data):
        return self.model.wrapper.call_method(self.method_name, data)


@make_string('id', 'name')
class Model(_InTask):
    """
    Model contains metadata for machine learning model

    :param name: model name
    :param wrapper_meta: :class:`~algolink.core.objects.wrapper.ModelWrapper` instance for this model
    :param artifact: :class:`~algolink.core.objects.ArtifactCollection` instance with model artifacts
    :param requirements: :class:`~algolink.core.objects.Requirements` instance with model requirements
    :param params: dict with arbitrary parameters. Must be json-serializable
    :param description: text description of this model
    :param id: model id
    :param task_id: parent task_id
    :param author: user that created that model
    :param creation_date: date when this model was created
    """

    PYTHON_VERSION = 'python_version'

    def __init__(self, name: str, wrapper_meta: Optional[dict] = None,
                 artifact: 'ArtifactCollection' = None,
                 requirements: Requirements = None,
                 params: Dict[str, Any] = None,
                 description: str = None,
                 id: int = None,
                 task_id: int = None,
                 author: str = None, creation_date: datetime.datetime = None,
                 evaluations: Dict[str, EvaluationResults] = None):
        super().__init__(id, name, author, creation_date, task_id)

        self.evaluations = evaluations or {}
        self.description = description
        self.params = params or {}
        try:
            json.dumps(self.params)
        except TypeError:
            raise ValueError('"params" argument must be json-serializable')
        self._wrapper = None
        self._wrapper_meta = None
        if isinstance(wrapper_meta, ModelWrapper):
            self._wrapper = wrapper_meta
        elif isinstance(wrapper_meta, dict):
            self._wrapper_meta = wrapper_meta

        self.requirements = requirements
        self._persisted_artifacts = artifact
        self._unpersisted_artifacts: Optional[ArtifactCollection] = None

    def load(self):
        """
        Load model artifacts into wrapper
        """
        if get_python_version() != self.params.get(self.PYTHON_VERSION):
            warnings.warn(f'Loading model from different python version {self.params.get(self.PYTHON_VERSION)}')
        with tempfile.TemporaryDirectory(prefix='algolink_run_') as tmpdir:
            self.artifact.materialize(tmpdir)
            self.wrapper.load(tmpdir)

    def ensure_loaded(self):
        """
        Ensure that wrapper has loaded model object
        """
        if self.wrapper.model is None:
            self.load()

    @property
    def wrapper(self) -> 'ModelWrapper':
        if self._wrapper is None:
            if self._wrapper_meta is None:
                raise ValueError("Either 'wrapper' or 'wrapper_meta' should be provided")
            self._wrapper = deserialize(self._wrapper_meta, ModelWrapper)
        return self._wrapper

    @wrapper.setter
    def wrapper(self, wrapper: ModelWrapper):
        if self._wrapper_meta is not None:
            raise ValueError("'wrapper' could be provided for models with no 'wrapper_meta' specified only")
        self._wrapper = wrapper

    def with_wrapper(self, wrapper: ModelWrapper):
        """
        Bind wrapper instance to this Model

        :param wrapper: :class:`~algolink.core.objects.wrapper.ModelWrapper` instance
        :return: self
        """
        self.wrapper = wrapper
        return self

    @property
    def wrapper_meta(self) -> dict:
        """
        :return: pyjackson representation of :class:`~algolink.core.objects.wrapper.ModelWrapper` for this model: e.g.,
          this provides possibility to move a model between repositories without its dependencies being installed
        """
        if self._wrapper_meta is None:
            if self._wrapper is None:
                raise ValueError("Either 'wrapper' or 'wrapper_meta' should be provided")
            self._wrapper_meta = serialize(self._wrapper)
        return self._wrapper_meta

    @wrapper_meta.setter
    def wrapper_meta(self, meta: dict):
        if self._wrapper is not None:
            raise ValueError("'wrapper_meta' could be provided for models with no 'wrapper' specified only")
        self._wrapper_meta = meta

    def with_wrapper_meta(self, wrapper_meta: dict):
        """
        Bind wrapper_meta dict to this Model

        :param wrapper_meta: dict with serialized :class:`~algolink.core.objects.wrapper.ModelWrapper` instance
        :return: self
        """
        self.wrapper_meta = wrapper_meta
        return self

    # this property is needed for pyjackson to serialize model, it is coupled with __init__
    @property
    def artifact(self) -> 'ArtifactCollection':
        """
        :return: persisted artifacts if any
        """
        return self._persisted_artifacts

    @property
    def artifact_any(self) -> 'ArtifactCollection':
        """
        :return: artifacts in any state (persisted or not)
        """
        arts = [a for a in [self._persisted_artifacts, self._unpersisted_artifacts] if a is not None]
        return CompositeArtifactCollection(arts) if len(arts) != 1 else arts[0]

    @property
    def artifact_req_persisted(self) -> 'ArtifactCollection':
        """
        Similar to `artifact` but checks that no unpersisted artifacts are left

        :return: persisted artifacts if any
        """
        if self._unpersisted_artifacts is not None:
            raise ValueError('Model has unpersisted artifacts')
        return self._persisted_artifacts

    def attach_artifact(self, artifact: 'ArtifactCollection'):
        """
        :param artifact: artifacts to attach to model in an unpersisted state
        """
        if self._unpersisted_artifacts is not None:
            self._unpersisted_artifacts += artifact
        else:
            self._unpersisted_artifacts = artifact

    def persist_artifacts(self, persister: Callable[['ArtifactCollection'], 'ArtifactCollection']):
        """
        Model artifacts persisting workflow

        :param persister: external object which stores model artifacts
        """
        artifact = self._persisted_artifacts

        if self._unpersisted_artifacts is None:
            if artifact is None:
                raise ValueError('Model has no artifacts')
        else:
            if artifact is None:
                artifact = self._unpersisted_artifacts
            else:
                artifact += self._unpersisted_artifacts

        self._persisted_artifacts = persister(artifact)
        self._unpersisted_artifacts = None

    def without_artifacts(self) -> 'Model':
        """
        :return: copy of the model with no artifacts attached
        """
        no_artifacts = copy(self)
        no_artifacts._persisted_artifacts = None
        no_artifacts._unpersisted_artifacts = None
        return no_artifacts

    @classmethod
    def create(cls, model_object, input_data, model_name: str = None,
               params: Dict[str, Any] = None, description: str = None,
               additional_artifacts: ArtifactCollection = None, additional_requirements: AnyRequirements = None,
               custom_wrapper: ModelWrapper = None, custom_artifact: ArtifactCollection = None,
               custom_requirements: AnyRequirements = None) -> 'Model':
        """
        Creates Model instance from arbitrary model objects and sample of input data

        :param model_object: The model object to analyze.
        :param input_data: Input data sample to determine structure of inputs and outputs for given model object.
        :param model_name: The model name.
        :param params: dict with arbitrary parameters. Must be json-serializable
        :param description: text description of this model
        :param additional_artifacts: Additional artifact.
        :param additional_requirements: Additional requirements.
        :param custom_wrapper: Custom model wrapper.
        :param custom_artifact: Custom artifact collection to replace all other.
        :param custom_requirements: Custom requirements to replace all other.
        :returns: :py:class:`Model`
        """
        wrapper: ModelWrapper = custom_wrapper or ModelAnalyzer.analyze(model_object, input_data=input_data)
        name = model_name or _generate_model_name(wrapper)

        artifact = custom_artifact or WrapperArtifactCollection(wrapper)
        if additional_artifacts is not None:
            artifact += additional_artifacts

        if custom_requirements is not None:
            requirements = resolve_requirements(custom_requirements)
        else:
            requirements = wrapper.requirements

        if additional_requirements is not None:
            requirements += additional_requirements

        requirements = RequirementAnalyzer.analyze(requirements)
        params = params or {}
        params[cls.PYTHON_VERSION] = params.get(cls.PYTHON_VERSION, get_python_version())
        # noinspection PyTypeChecker
        model = Model(name, wrapper, None, requirements, params, description)
        model._unpersisted_artifacts = artifact
        return model

    @_with_meta
    @_with_artifact
    @ExposedObjectMethod('delete_model', 'model', 'Model', 'model to delete')
    def delete(self, force: bool = False):
        """
        Deletes model from metadata and artifact repositories

        :param force: whether model artifacts' deletion errors should be ignored, default is false
        :return: Nothing
        """
        if self.artifact is not None:
            try:
                self._art.delete_model_artifact(self)
            except:  # noqa
                if force:
                    logger.warning("Unable to delete artifacts associated with model: '%s'", self, exc_info=1)
                else:
                    raise

        self._meta.delete_model(self)
        self.task_id = None

    @_with_meta(saved=False)
    @_with_artifact
    def push(self, task: Task = None) -> 'Model':
        """
        Pushes :py:class:`~algolink.core.objects.Model` instance into metadata and artifact repositories

        :param task: :py:class:`~algolink.core.objects.Task` instance to save model to. Optional if model already has
        task
        :return: same saved :py:class:`~algolink.core.objects.Model` instance
        """
        if self.id is not None:
            raise errors.ExistingModelError(self)
        if task is not None:
            if self.task_id is not None:
                if self.task_id != task.id:
                    raise ValueError('This model is already in task {}'.format(self.task_id))
            else:
                self.task = task

        self._meta.create_model(self)  # save model to get model.id
        try:
            self._art.push_model_artifacts(self)
        except:  # noqa
            self._meta.delete_model(self)
            raise

        return self._meta.save_model(self)

    def as_pipeline(self, method_name=None) -> 'Pipeline':
        """Create Pipeline that consists of this model's single method

        :param method_name: name of the method. can be omitted if model has only one method
        """
        method_name = self.wrapper.resolve_method(method_name)
        method = self.wrapper.methods[method_name]
        pipeline = Pipeline(f'{self.name}.{method_name}',
                            [], method[1], method[2], task_id=self.task_id).append(self, method_name)
        return pipeline

    def __getattr__(self, item: str):
        if item.startswith('__') and item.endswith('__') and item != '__call__':
            # no special dunder attributes
            raise AttributeError()
        return _WrapperMethodAccessor(self, item)

    @_with_meta
    @_with_artifact
    def save(self):
        """Saves model to metadata repo and pushes unpersisted artifacts"""
        if self._unpersisted_artifacts is not None:
            self._art.push_model_artifacts(self)
        self._meta.save_model(self)

    def has_children(self):
        return False

    @_with_meta
    def evaluate_set(self, evalset: Union[str, EvaluationSet],
                     evaluation_name: str = None, method_name: str = None,
                     timestamp=None, save=True, force=False,
                     raise_on_error=False) -> Optional[EvaluationResult]:
        """Evaluates this model

        :param evalset: evalset or it's name
        :param evaluation_name: name of this evaluation
        :param method_name: name of wrapper method. If none, all methods with consistent datatypes will be evaluated
        :param timestamp: time of the evaluation (defaults to now)
        :param save: save results to meta
        :param force: force reevalute
        :param raise_on_error: raise error if datatypes are incorrect or just return
        """
        task = self.task
        if isinstance(evalset, str):
            try:
                evaluation_name = evaluation_name or evalset
                evalset = task.evaluation_sets[evalset]
            except KeyError:
                raise ValueError(f'No evalset {evalset} in {task}')
        input, output, metrics = evalset.get(task, False)
        return self.evaluate(input, output, metrics, evaluation_name, method_name, timestamp, save, force,
                             raise_on_error)

    def evaluate(self, input: DatasetSource, output: DatasetSource, metrics: Dict[str, Metric],
                 evaluation_name: str = None, method_name: str = None,
                 timestamp=None, save=True, force=False,
                 raise_on_error=False) -> Optional[Union[EvaluationResult, Dict[str, EvaluationResult]]]:
        """Evaluates this model

        :param input: input data
        :param output: target
        :param metrics: dict of metrics to evaluate
        :param evaluation_name: name of this evaluation
        :param method_name: name of wrapper method. If none, all methods with consistent datatypes will be evaluated
        :param timestamp: time of the evaluation (defaults to now)
        :param save: save results to meta
        :param force: force reevalute
        :param raise_on_error: raise error if datatypes are incorrect or just return
        """
        if method_name is None:
            methods = self.wrapper.match_methods_by_type(input.dataset_type, output.dataset_type)
            if len(methods) == 0:
                if raise_on_error:
                    raise ValueError('incompatible dataset types for evaluation')
                return
        else:
            method_name = self.wrapper.resolve_method(method_name)
            tin, tout = self.wrapper.method_signature(method_name)
            if tin != input.dataset_type or tout != output.dataset_type:
                if raise_on_error:
                    raise ValueError('incompatible dataset types for evaluation')
                return
            methods = [method_name]

        if save:
            if evaluation_name is None:
                raise ValueError('Provide evaluation_name to save evaluation or set save = False')
            self._check_meta(True)
        results: Dict[str, EvaluationResult] = {}  # method -> result
        timestamp = timestamp or time.time()

        for method_name in methods:
            self.evaluations.setdefault(method_name, {})
            if evaluation_name in self.evaluations[method_name] and not force:
                results[method_name] = self.evaluations[method_name][evaluation_name].latest
                continue

            call = self.wrapper.call_method(method_name, input.read().data)
            scores = {}
            for mname, metric in metrics.items():
                scores[mname] = metric.evaluate(output.read().data, call)
            results[method_name] = EvaluationResult(timestamp, scores)

        if save:
            for method_name, result in results.items():
                self.evaluations[method_name].setdefault(evaluation_name, EvaluationResultCollection())
                self.evaluations[method_name][evaluation_name].add(result)
            self.save()

        if len(results) == 1:
            return list(results.values())[0]
        return results


def _generate_name(prefix='', postfix=''):
    """Generates name from current date

    :param prefix: str to put before
    :param postfix: str to put after"""
    now = datetime.datetime.now()
    return f'{prefix}{now.strftime("%Y%m%d_%H_%M_%S_%f")}{postfix}'


def _generate_model_name(wrapper: ModelWrapper):
    """
    Generates name for Model instance

    :param wrapper: model wrapper
    :return: str
    """

    return _generate_name('{}_model_'.format(wrapper.type))


class PipelineStep(AlgoLinkParams):
    """A class to represent one step of a Pipeline - a Model with one of its' methods name

    :param model_name: name of the Model (in the same Task as Pipeline object)
    :param method_name: name of the method in Model's wrapper to use"""

    def __init__(self, model_name: str, method_name: str):
        self.model_name = model_name
        self.method_name = method_name


@make_string('id', 'name')
class Pipeline(_InTask):
    """Pipeline is a class to represent a sequence of different Model's methods.
    They can be used to reuse different models (for example, pre-processing functions) in different pipelines.
    Pipelines must have exact same in and out data types as tasks they are in

    :param name: name of the pipeline
    :param steps: sequence of :class:`.PipelineStep`s the pipeline consists of
    :param input_data: datatype of input dataset
    :param output_data: datatype of output datset
    :param id: id of the pipeline
    :param author: author of the pipeline
    :param creation_date: date of creation
    :param task_id: task.id of parent task"""

    def __init__(self, name: str,
                 steps: List[PipelineStep],
                 input_data: DatasetType,
                 output_data: DatasetType,
                 id: int = None,
                 author: str = None, creation_date: datetime.datetime = None,
                 task_id: int = None,
                 evaluations: EvaluationResults = None):
        super().__init__(id, name, author, creation_date, task_id)
        self.evaluations = evaluations or {}
        self.output_data = output_data
        self.input_data = input_data
        self.steps = steps
        self.models: Dict[str, Model] = {}  # not using direct fk to models as it is pain

    @_with_meta
    @ExposedObjectMethod('delete_pipeline', 'pipeline', 'Pipeline', 'pipeline to delete')
    def delete(self):
        """Deletes pipeline from metadata"""
        self._meta.delete_pipeline(self)

    @_with_meta
    def load(self):
        task = self.task
        for step in self.steps:
            model = task.models(step.model_name)
            self.models[model.name] = model

    def run(self, data):
        """Applies sequence of pipeline steps to data

        :param data: data to apply pipeline to. must have type `Pipeline.input_data`
        :returns: processed data of type `Pipeline.output_data`"""
        for step in self.steps:
            model = self.models[step.model_name]
            data = model.wrapper.call_method(step.method_name, data)
        return data

    def append(self, model: Union[Model, _WrapperMethodAccessor], method_name: str = None):
        """Appends another Model to the sequence of this pipeline steps

        :param model: either Model instance, or model method (as in `model.method` where `method` is method name)
        :param method_name: if Model was provided in `model`, this should be method name.
        can be omitted if model have only one method"""
        # TODO datatype validaiton
        if isinstance(model, _WrapperMethodAccessor):
            method_name = model.method_name
            model = model.model
        method_name = model.wrapper.resolve_method(method_name)

        self.steps.append(PipelineStep(model.name, method_name))
        self.models[model.name] = model
        self.output_data = model.wrapper.methods[method_name][2]  # TODO change it to namedtuple
        return self

    @_with_meta
    def save(self):
        """Saves this pipeline to metadata repository"""
        self._meta.save_pipeline(self)

    def has_children(self):
        return False

    @_with_meta
    def evaluate_set(self, evalset: Union[str, EvaluationSet],
                     evaluation_name: str = None,
                     timestamp=None, save=True, force=False,
                     raise_on_error=False) -> Optional[EvaluationResult]:
        """Evaluates this pipeline

        :param evalset: evalset or it's name
        :param evaluation_name: name of this evaluation
        :param timestamp: time of the evaluation (defaults to now)
        :param save: save results to meta
        :param force: force reevalute
        :param raise_on_error: raise error if datatypes are incorrect or just return
        """
        task = self.task
        if isinstance(evalset, str):
            try:
                evaluation_name = evaluation_name or evalset
                evalset = task.evaluation_sets[evalset]
            except KeyError:
                raise ValueError(f'No evalset {evalset} in {task}')
        input, output, metrics = evalset.get(task, False)
        return self.evaluate(input, output, metrics, evaluation_name, timestamp, save, force, raise_on_error)

    def evaluate(self, input: DatasetSource, output: DatasetSource, metrics: Dict[str, Metric],
                 evaluation_name: str = None,
                 timestamp=None, save=True, force=False,
                 raise_on_error=False) -> Optional[EvaluationResult]:
        """Evaluates this pipeline

        :param input: input data
        :param output: target
        :param metrics: dict of metrics to evaluate
        :param evaluation_name: name of this evaluation
        :param timestamp: time of the evaluation (defaults to now)
        :param save: save results to meta
        :param force: force reevalute
        :param raise_on_error: raise error if datatypes are incorrect or just return
        """
        if evaluation_name in self.evaluations and not force:
            return self.evaluations[evaluation_name].latest
        if save:
            if evaluation_name is None:
                raise ValueError('Provide evaluation_name to save evaluation or set save = False')
            self._check_meta(True)
        timestamp = timestamp or time.time()

        if input.dataset_type == self.input_data and output.dataset_type == self.output_data:
            scores = {}
            call = self.run(input.read().data)
            for mname, metric in metrics.items():
                scores[mname] = metric.evaluate(output.read().data, call)
            result = EvaluationResult(timestamp, scores)
            if save:
                self.evaluations.setdefault(evaluation_name, EvaluationResultCollection())
                self.evaluations[evaluation_name].add(result)
                self.save()
            return result
        elif raise_on_error:
            raise ValueError('incompatible dataset types for evaluation')


@type_field('type')
class Buildable(AlgoLinkParams, WithMetadataRepository):
    """An abstract class that represents something that can be built by Builders
    Have default implementations for Models and Pipelines (and lists of them)
    """

    @abstractmethod
    def get_provider(self):
        """Abstract method to get a provider for this Buildable"""

    @property
    @abstractmethod
    def task(self) -> Optional[Task]:
        """property to get task (can be None, whick forces to provide task manually)"""


class RuntimeEnvironment(AlgoLinkObject):
    """Represents and environment where you can build and deploy your services
    Actual type of environment depends on `.params` field type

    :param name: name of the environment
    :param id: id of the environment
    :param author: author of the enviroment
    :param creation_date: creation date of the enviroment
    :param params: :class:`.RuntimeEnvironment.Params` instance
    """

    @type_field('type')
    class Params(AlgoLinkParams):
        """Abstract class that represents different types of environments"""
        default_runner = None
        default_builder = None

        def get_runner(self):
            """
            :return: Runner for this environment
            """
            return self.default_runner

        def get_builder(self):
            """
            :return: builder for this environment
            """
            return self.default_builder

    def __init__(self, name: str, id: int = None, params: Params = None,
                 author: str = None, creation_date: datetime.datetime = None):
        super().__init__(id, name, author, creation_date)
        self.params = params

    @_with_meta
    @ExposedObjectMethod('delete_environment', 'environment', 'RuntimeEnvironment', 'environment to delete')
    def delete(self, meta_only: bool = False, cascade: bool = False):
        """
        Deletes environment from metadata repository and(if required) stops associated instances

        :param meta_only: wheter to only delete metadata
        :param cascade: Whether should environment be deleted with all associated instances
        :return: Nothing
        """
        if cascade:
            instances = self._meta.get_instances(image=None, environment=self)
            for instance in instances:
                instance.delete(meta_only=meta_only)
        self._meta.delete_environment(self)

    @_with_meta
    def save(self):
        """Saves this env to metadata repository"""
        self._meta.save_environment(self)

    @_with_meta
    def has_children(self):
        return len(self._meta.get_instances(image=None, environment=self)) > 0


class _WithEnvironment(AlgoLinkObject):
    """Utility class for objects with PK to :class:`.RuntimeEnvironment`"""

    def __init__(self, id: int, name: str, author: str = None, creation_date: datetime.datetime = None,
                 environment_id: int = None):
        super().__init__(id, name, author, creation_date)
        self.environment_id = environment_id
        self._environment: Optional[RuntimeEnvironment] = None

    @property
    def environment(self) -> RuntimeEnvironment:  # TODO caching
        if self._environment is not None:
            return self._environment
        if self.environment_id is None:
            raise errors.UnboundObjectError(f"{self} is not saved, cant access environment")
        self._check_meta()
        e = self._meta.get_environment_by_id(self.environment_id)
        if e is None:
            raise errors.NonExistingEnvironmentError(self.environment_id)
        e = e.bind_as(self)
        self._environment = e
        return self._environment

    @environment.setter
    def environment(self, environment: RuntimeEnvironment):
        if not isinstance(environment, RuntimeEnvironment):
            raise ValueError(f'{environment} is not RuntimeEnvironment')
        self.environment_id = environment.id
        self._environment = environment
        self.bind_as(environment)


class _WithBuilder(_WithEnvironment):
    builder = None

    def bind_builder(self, builder):
        self.builder = builder
        return self

    def unbind_builder(self):
        del self.builder

    @property
    def has_builder(self):
        return self.builder is not None


def _with_auto_builder(method):
    """
    Decorator for methods to check that object is binded to builder

    :param method: method to apply decorator
    :return: decorated method
    """

    @wraps(method)
    def inner(self: _WithBuilder, *args, **kwargs):
        if not self.has_builder:
            if not self.has_meta_repo:
                raise ValueError(f'{self} has no binded runner')
            self.bind_builder(self.environment.params.get_builder())
        return method(self, *args, **kwargs)

    return inner


@make_string('id', 'name')
class Image(_WithBuilder):
    """Class that represents metadata for image built from Buildable
    Actual type of image depends on `.params` field type

    :param name: name of the image
    :param id: id of the image
    :param author: author of the image
    :parma creation_date: creation date of the image
    :param source: :class:`.Buildable` instance this image was built from
    :param params: :class:`.Image.Params` instance
    :param task_id: task.id this image belongs to
    :param environment_id: environment.id this image belongs to
    """

    @type_field('type')
    class Params(AlgoLinkParams):
        """Abstract class that represents different types of images"""

    def __init__(self, name: Optional[str], source: Buildable, id: int = None,
                 params: Params = None,
                 author: str = None, creation_date: datetime.datetime = None,
                 task_id: int = None,
                 environment_id: int = None):
        super().__init__(id, name or _generate_name('ebnt_image_'), author, creation_date, environment_id)
        self.task_id = task_id
        self.source = source
        self.params = params

    @property
    @_with_meta
    def task(self):
        t = self._meta.get_task_by_id(self.task_id)
        if t is None:
            raise errors.NonExistingTaskError(self.task_id)
        return t.bind_artifact_repo(self._art)

    @task.setter
    def task(self, task: Task):
        if not isinstance(task, Task):
            raise ValueError('{} is not Task'.format(task))
        self.task_id = task.id
        self.bind_as(task)

    @_with_meta
    @ExposedObjectMethod('delete_image', 'image', 'Image', 'image ot delete')
    def delete(self, meta_only: bool = False, cascade: bool = False):
        """
        Deletes existing image from metadata repository and image provider

        :param meta_only: should image be deleted only from metadata
        :param cascade: whether to delete nested RuntimeInstances
        """
        if cascade:
            for instance in self._meta.get_instances(self):
                instance.delete(meta_only=meta_only)
        elif len(self._meta.get_instances(self)) > 0:
            raise errors.ImageWithInstancesError(self)

        if not meta_only:
            self.remove()
        self._meta.delete_image(self)

    def bind_meta_repo(self, repo: 'algolink.repository.MetadataRepository'):
        super(Image, self).bind_meta_repo(repo)
        self.source.bind_meta_repo(repo)
        return self

    @_with_auto_builder
    def is_built(self) -> bool:
        """Checks if image was built and wasn't removed"""
        return self.builder.image_exists(self.params, self.environment.params)

    @_with_auto_builder
    def build(self, **kwargs):
        """Build this image

        :param kwargs: additional params for builder.build_image (depends on builder implementation)
        """
        self.builder.build_image(self.source, self.params, self.environment.params, **kwargs)
        if self.has_meta_repo:
            self._meta.save_image(self)
        return self

    @_with_auto_builder
    def remove(self, **kwargs):
        """remove this image (from environment, not from algolink metadata)"""
        self.builder.delete_image(self.params, self.environment.params, **kwargs)

    @_with_meta
    def save(self):
        self._meta.save_image(self)

    def has_children(self):
        return False


class _WithRunner(_WithEnvironment):
    runner = None

    def bind_runner(self, runner):
        self.runner = runner
        return self

    def unbind_runner(self):
        del self.runner

    @property
    def has_runner(self):
        return self.runner is not None


def _with_auto_runner(method):
    """
    Decorator for methods to check that object is binded to runner

    :param method: method to apply decorator
    :return: decorated method
    """

    @wraps(method)
    def inner(self: '_WithRunner', *args, **kwargs):
        if not self.has_runner:
            if not self.has_meta_repo:
                raise ValueError(f'{self} has no binded runner')
            self.bind_runner(self.environment.params.get_runner())
        return method(self, *args, **kwargs)

    return inner


class RuntimeInstance(_WithRunner):
    """Class that represents metadata for instance running in environment
    Actual type of instance depends on `.params` field type

    :param name: name of the instance
    :param id: id of the instance
    :param author: author of the instance
    :parma creation_date: creation date of the instance
    :param image_id: id of base image for htis instance
    :param params: :class:`.RuntimeInstance.Params` instance
    """

    @type_field('type')
    class Params(AlgoLinkParams):
        """Abstract class that represents different types of images"""

    def __init__(self, name: Optional[str], id: int = None,
                 image_id: int = None, environment_id: int = None, params: Params = None,
                 author: str = None, creation_date: datetime.datetime = None):
        super().__init__(id, name or _generate_name('ebnt_instance_'), author, creation_date, environment_id)
        self.image_id = image_id
        self.params = params

    @property
    @_with_meta
    def image(self) -> Image:
        i = self._meta.get_image_by_id(self.image_id)
        if i is None:
            raise errors.NonExistingImageError(self.image_id)
        return i.bind_artifact_repo(self._art)

    @image.setter
    def image(self, image: Image):
        if not isinstance(image, Image):
            raise ValueError(f'{image} is not Image')
        self.image_id = image.id
        self.bind_as(image)

    @_with_meta
    @ExposedObjectMethod('delete_instance', 'instance', 'RuntimeInstance', 'instance to delete')
    def delete(self, meta_only: bool = False):
        """
        Stops instance of model service and deletes it from repository

        :param meta_only: only remove from metadata, do not stop instance
        :return: nothing
        """
        if not meta_only:
            self.stop()
            self.remove()

        self._meta.delete_instance(self)

    @_with_auto_runner
    def run(self, **runner_kwargs) -> 'RuntimeInstance':
        """Run this instance

        :param runner_kwargs: additional params for runner.run (depends on runner implementation)
        """
        self.runner.run(self.params, self.image.params, self.environment.params, **runner_kwargs)
        if self.has_meta_repo:
            self._meta.save_instance(self)
        return self

    @_with_auto_runner
    def logs(self, **kwargs):
        """Get logs of this instance

        :param kwargs: parameters for runner `logs` method
        :yields: str logs from running instance
        """
        yield from self.runner.logs(self.params, self.environment.params, **kwargs)

    @_with_auto_runner
    def is_running(self, **kwargs) -> bool:
        """
        Checks whether instance is running

        :param kwargs: params for runner `is_running` method
        :return: "is running" flag
        """
        return self.runner.is_running(self.params, self.environment.params, **kwargs)

    @_with_auto_runner
    def stop(self, **kwargs):
        """Stops the instance

        :param kwargs: params for runner `stop` method
        """
        self.runner.stop(self.params, self.environment.params, **kwargs)

    @_with_auto_runner
    def exists(self, **kwargs) -> bool:
        """Checks if instance exists (it may be stopped)

        :param kwargs: params for runner `instance_exists` method"""
        return self.runner.instance_exists(self.params, self.environment.params, **kwargs)

    @_with_auto_runner
    def remove(self, **kwargs):
        """Removes the instance from environment (not from metadata)

         :param kwargs: params for runner `remove_instance` method"""
        self.runner.remove_instance(self.params, self.environment.params, **kwargs)

    @_with_meta
    def save(self):
        self._meta.save_instance(self)

    def has_children(self):
        return False
    
## ML. tracking leepand
@make_string('id', 'name')
class Experiment(AlgoLinkObject,WithDatasetRepository):
    """
    Experiment is a collection of submodels

    :param id: experiment id
    :param name: experiment name
    :param author: user that created that experiment
    :param creation_date: date when this experiment was created
    """
    #_nested_fields_meta = _nested_fields_art = ['_modelmetrics,','_modelparams','_bestresult']

    def __init__(self, name: str, 
                 id: int = None, 
                 author: str = None, 
                 experiment_remark: str = None,
                 experiment_sequence: int = None,
                 del_flag: int = None,
                 task_id: int = None,
                 creation_date: datetime.datetime = None):
        super().__init__(id, name, author, creation_date)
        self.experiment_remark="experiment_remark"
        self.experiment_sequence = experiment_sequence
        self.task_id = task_id
        self.del_flag = del_flag
        """self._modelmetrics: IndexDict[ModelMetric] = IndexDict('id', 'name')
        self.modelmetrics: IndexDictAccessor[ModelMetric] = IndexDictAccessor(self._modelmetrics)
        print(self._modelmetrics,"self._modelmetrics")
        self._modelparams: IndexDict[ModelParam] = IndexDict('id', 'name')
        self.modelparams: IndexDictAccessor[ModelParam] = IndexDictAccessor(self._modelparams)
        self._bestresult: IndexDict[BestResult] = IndexDict('id', 'name')
        self.bestresult: IndexDictAccessor[BestResult] = IndexDictAccessor(self._bestresult)"""

    @_with_meta
    @ExposedObjectMethod(name='delete_experiment', param_name='experiment', param_type='Experiment',
                         param_doc='experiment to delete')
    def delete(self, cascade: bool = False):
        """
        Deletes experiment and(if required) all submodels associated with it from metadata repository

        :param cascade: whether should project be deleted with all associated submodels
        :return: Nothing
        """
        if cascade:
            for submodel in self._meta.get_experiments(self):  # TODO cacheing?
                self.delete_experiment(experiment, cascade=cascade)
                # task.delete(cascade=cascade) # FIXME
        self._meta.delete_task(self)

    @_with_meta
    def add_metric(self, metric_name: str,metric_value: float,epoch: int):
        """
        Add model metrics to experiment and save it to meta repo

        :param experiment: experiment to add
        """
        metric = ModelMetric(name=metric_name,metric_type = "metric_type",metric_value = metric_value,epoch=epoch, experiment_id = self.id)
        
        if metric.experiment_id is not None and metric.experiment_id != self.id:
            raise errors.MetadataError('Metric is already in experiment {}. Delete it first'.format(metric.experiment_id))

        metric.experiment_id = self.id
        ## 持久化至数据库
        self._meta.save_metric(metric)
        #self._modelmatrics.add(metric)
        return metric.bind_as(self)

    @_with_meta
    def add_param(self, param_dict: dict,param_type: str):
        """
        Add model params to experiments and save them to meta repo

        :param param_dict: dict，训练参数字典
        :param param_type: str，参数类型，例如：TF参数、word2vec参数等。
        """
        _param_dict = {}
        _param_dict[param_type] = param_dict
        # 插入model param
        for param_type, value in _param_dict.items():
            for param_name, param_value in value.items():
                param_data = ModelParam(name =param_name,param_type=param_type, 
                                        param_value=str(param_value),experiment_id=self.id)
                if param_data.experiment_id is not None and param_data.experiment_id != self.id:
                    raise errors.MetadataError('param_data is already in experiment {}. Delete it first'.format(param_data.experiment_id))
                param_data.experiment_id = self.id
                ## 持久化至数据库
                self._meta.save_param(param_data)
            return param_data.bind_as(self)

    @_with_meta
    def add_best_result(self, best_name: str, best_value: float,best_epoch:int):
        """
        Add model best param to experiments and save them to meta repo

        :param best_name: best param's name
        :param best_value: best value
        :param best_epoch: best epoch
        """
        best_result = BestResult(name = best_name,best_value = best_value,best_epoch=best_epoch)
        best_result.experiment_id = self.id
        ## 持久化至数据库
        self._meta.save_best_result(best_result)
        #self._modelmatrics.add(metric)
        return best_result.bind_as(self)

    def __repr__(self):
        return """Task '{name}', {td} experiments""".format(name=self.name, td=self.experiment_remark)
    
@make_string('id', 'name')
class ModelMetric(AlgoLinkObject):
    """
    ModelMetric contains metadata for machine learning model's metrics

    :param name: metric name
    """

    PYTHON_VERSION = 'python_version'

    def __init__(self, name: str, 
                 metric_type: str = None,
                 metric_value: float = None,
                 epoch: int = None,
                 id: int = None,
                 experiment_id: int = None,
                 author: str = None, 
                 creation_date: datetime.datetime = None):
        super().__init__(id, name, author, creation_date)
        self.metric_type = metric_type
        self.metric_value = metric_value
        self.epoch =epoch
        self.experiment_id = experiment_id
        
        
@make_string('id', 'name')
class ModelParam(AlgoLinkObject):
    """
    ModelParam contains metadata for machine learning model's params

    :param name: param name
    """

    PYTHON_VERSION = 'python_version'

    def __init__(self, name: str, 
                 param_type: str = None,
                 param_value: float = None,
                 id: int = None,
                 experiment_id: int = None,
                 author: str = None, 
                 creation_date: datetime.datetime = None):
        super().__init__(id, name, author, creation_date)
        self.param_type = param_type
        self.param_value = param_value
        self.experiment_id = experiment_id      

@make_string('id', 'name')
class BestResult(AlgoLinkObject):
    """
    BestResult contains metadata for machine learning model's best result

    :param name: result name
    """

    PYTHON_VERSION = 'python_version'

    def __init__(self, name: str, 
                 best_epoch: int = None,
                 best_value: float = None,
                 id: int = None,
                 experiment_id: int = None,
                 author: str = None, 
                 creation_date: datetime.datetime = None):
        super().__init__(id, name, author, creation_date)
        self.best_epoch = best_epoch
        self.best_value = best_value
        self.experiment_id = experiment_id


import re
from typing import List, Union

import numpy as np
import pandas as pd
from pandas import Int64Dtype, SparseDtype, StringDtype
from pandas.core.dtypes.dtypes import (CategoricalDtype, DatetimeTZDtype, IntervalDtype, PandasExtensionDtype,
                                       PeriodDtype)
from pyjackson.core import ArgList, Field
from pyjackson.decorators import cached_property
from pyjackson.errors import DeserializationError, SerializationError

from algolink.core.analyzer.base import TypeHookMixin
from algolink.core.analyzer.dataset import DatasetHook
from algolink.core.objects.dataset_type import DatasetType, LibDatasetTypeMixin
from algolink.ext.numpy.dataset import np_type_from_string, python_type_from_np_type

_PD_EXT_TYPES = {
    DatetimeTZDtype: r'datetime64.*',
    CategoricalDtype: r'category',
    PeriodDtype: r'[pP]eriod.*',
    SparseDtype: r'Sparse.*',
    IntervalDtype: r'[iI]nterval.*',
    Int64Dtype: r'U?Int\d*',
    StringDtype: r'string'
}
PD_EXT_TYPES = {dtype: re.compile(pattern) for dtype, pattern in
                _PD_EXT_TYPES.items()}


def string_repr_from_pd_type(dtype: Union[np.dtype, PandasExtensionDtype]) -> str:
    """Returns string representation of pandas dtype"""
    return dtype.name


def pd_type_from_string(string_repr):
    """Creates pandas dtype from string representation"""
    try:
        return np_type_from_string(string_repr)
    except ValueError:
        for dtype, pattern in PD_EXT_TYPES.items():
            if pattern.match(string_repr) is not None:
                return dtype.construct_from_string(string_repr)
        raise ValueError(f'unknown pandas dtype {string_repr}')


def python_type_from_pd_type(pd_type: np.dtype):
    """Returns python builtin type that corresponds to pandas dtype"""
    if isinstance(pd_type, np.dtype):  # or (isinstance(pd_type, type) and pd_type.__module__ == 'numpy'):
        return python_type_from_np_type(pd_type)
    return str


def python_type_from_pd_string_repr(string_repr: str) -> type:
    """Returns python type from pandas dtype string representation"""
    pd_type = pd_type_from_string(string_repr)
    return python_type_from_pd_type(pd_type)


def need_string_value(dtype):
    """Returns true if dtype must be cast to str for serialization"""
    return not isinstance(dtype, np.dtype) or dtype.type in (np.datetime64, np.object_)


class PandasDFHook(TypeHookMixin, DatasetHook):
    """
    :class:`.DatasetHook` implementation for `pandas.DataFrame` which uses :class:`DataFrameType`
    """

    valid_types = [pd.DataFrame]

    def process(self, obj: pd.DataFrame, **kwargs) -> DatasetType:
        if has_index(obj):
            index_cols, obj = _reset_index(obj)
        else:
            index_cols = []

        return DataFrameType(list(obj.columns), [string_repr_from_pd_type(d) for d in obj.dtypes], index_cols)


class _PandasDatasetType(LibDatasetTypeMixin):
    """Intermidiate class for pandas DatasetType implementations

    :param columns: list of column names (including index)
    :param dtypes: list of string representations of pandas dtypes of columns
    :param index_cols: list of column names that are used as index"""
    libraries = [pd]

    def __init__(self, columns: List[str], dtypes: List[str], index_cols: List[str]):
        self.index_cols = index_cols
        self.columns = columns
        self.dtypes = dtypes

    @cached_property
    def actual_dtypes(self):
        """List of pandas dtypes for columns"""
        return [pd_type_from_string(s) for s in self.dtypes]

    def _validate_columns(self, df: pd.DataFrame, exc_type):
        """Validates that df has correct columns"""
        if set(df.columns) != set(self.columns):
            raise exc_type(f'given dataframe has columns: {list(df.columns)}, expected: {self.columns}')

    def _validate_dtypes(self, df: pd.DataFrame, exc_type):
        """Validates that df has correct column dtypes"""
        df = df[self.columns]
        for col, expected, dtype in zip(self.columns, self.dtypes, df.dtypes):
            pd_type = string_repr_from_pd_type(dtype)
            if expected != pd_type:
                raise exc_type(f'given dataframe has incorrect type {pd_type} in column {col}. Expected: {expected}')


class SeriesType(_PandasDatasetType):
    """
    :class:`.DatasetType` implementation for `pandas.Series` objects which stores them as built-in Python dicts

    """
    real_type = pd.Series

    def deserialize(self, obj):
        return pd.Series(obj)

    def serialize(self, instance: pd.Series):
        return instance.to_dict()

    def get_spec(self):
        return [Field(c, python_type_from_pd_string_repr(d), False) for c, d in zip(self.columns, self.dtypes)]


def has_index(df: pd.DataFrame):
    """Returns true if df has non-trivial index"""
    return not isinstance(df.index, pd.RangeIndex)


def _reset_index(df: pd.DataFrame):
    """Transforms indexes to columns"""
    index_name = df.index.name or ''  # save it for future renaming
    cols = set(df.columns)
    df = df.reset_index()  # can rename indexes if they didnt have a name
    index_cols = [c for c in df.columns if c not in cols]  # get new columns - they were index columns
    if len(index_cols) == 1:
        df = df.rename({index_cols[0]: index_name}, axis=1)
        index_cols = [index_name]
    return index_cols, df


def reset_index(df: pd.DataFrame, return_copied=False):
    """Transforms indexes to columns if index is non-trivial"""
    if has_index(df):
        _, df = _reset_index(df)
        if return_copied:
            return True, df
        return df
    if return_copied:
        return False, df
    return df


class DataFrameType(_PandasDatasetType):
    """
    :class:`.DatasetType` implementation for `pandas.DataFrame`
    """

    real_type = pd.DataFrame

    def deserialize(self, obj):
        self._check_type(obj, dict, DeserializationError)
        try:
            ret = pd.DataFrame.from_records(obj['values'])
        except (ValueError, KeyError):
            raise DeserializationError(f'given object: {obj} could not be converted to dataframe')

        self._validate_columns(ret, DeserializationError)  # including index columns
        ret = self.align_types(ret)  # including index columns
        self._validate_dtypes(ret, DeserializationError)
        return self.align_index(ret)

    def align_types(self, df):
        """Restores column order and casts columns to expected types"""
        df = df[self.columns]
        for col, expected, dtype in zip(self.columns, self.actual_dtypes, df.dtypes):
            if expected != dtype:
                df[col] = df[col].astype(expected)
        return df

    def align_index(self, df):
        """Transform index columns to actual indexes"""
        if len(self.index_cols) > 0:
            df = df.set_index(self.index_cols)
        return df

    def align(self, df):
        return self.align_index(self.align_types(df))

    def serialize(self, instance: pd.DataFrame):
        self._check_type(instance, pd.DataFrame, SerializationError)
        is_copied, instance = reset_index(instance, return_copied=True)

        self._validate_columns(instance, SerializationError)
        self._validate_dtypes(instance, SerializationError)

        for col, dtype in zip(self.columns, self.actual_dtypes):
            if need_string_value(dtype):
                if not is_copied:
                    instance = instance.copy()
                    is_copied = True
                instance[col] = instance[col].astype('string')

        return {'values': (instance.to_dict('records'))}

    def get_spec(self) -> ArgList:
        return [Field('values', List[self.row_type], False)]

    @cached_property
    def row_type(self):
        return SeriesType(self.columns, self.dtypes, self.index_cols)

    def get_writer(self):
        from .dataset_source import PandasWriter, PandasFormatCsv
        return PandasWriter(PandasFormatCsv())  # TODO env configuration



import os
from typing import Optional, Union

from algolink.build.provider import PythonProvider
from algolink.build.provider.utils import BuildableWithServer
from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.buildable import BuildableHook
from algolink.core.objects.artifacts import _RelativePathWrapper, CompositeArtifactCollection, Blobs, LocalFileBlob
from algolink.core.objects import ArtifactCollection, Model, Requirements, Task
from pyjackson import dumps
from pyjackson.decorators import cached_property

from algolink.core.objects.core import WithMetadataRepository
from algolink.runtime.interface.ml_model import MODEL_BIN_PATH, MODEL_META_PATH
from algolink.runtime.server import Server
from algolink.utils.module import get_object_requirements

LOADER_PATH = 'loader'
SERVER_PATH = 'server'


def read(path):
    with open(path) as f:
        return f.read()


class MLModelProvider(PythonProvider):
    """Provider to build service from Model object

    :param model: Model instance to build from
    :param server: Server instance to build with
    :param debug: Whether to run image in debug mode
    """

    def __init__(self, model: Model, server: Server, debug: bool = False):
        from algolink.runtime.interface.ml_model import ModelLoader
        super().__init__(server, ModelLoader(), debug)
        self.model = model

    @cached_property
    def _requirements(self) -> Requirements:
        """Union of model, server and loader requirements"""
        return (self.model.requirements +
                get_object_requirements(self.server) +
                get_object_requirements(self.loader))

    def get_requirements(self) -> Requirements:
        """Returns union of model, server and loader requirements"""
        return self._requirements

    def _get_sources(self):
        """Returns sources of custom modules from requirements"""
        sources = {}
        for cr in self.get_requirements().custom:
            sources.update(cr.to_sources_dict())

        # add __init__.py for all dirs that doesnt have it already
        packages = set(os.path.join(os.path.dirname(p), '__init__.py') for p in sources if os.path.dirname(p) != '')
        sources.update({
            p: '' for p in packages if p not in sources
        })
        return sources

    def get_sources(self):
        """Returns model metadata file and sources of custom modules from requirements"""
        return {
            MODEL_META_PATH: dumps(self.model.without_artifacts()),
            **self._get_sources(),
            **{os.path.basename(f): read(f) for f in self.server.additional_sources}
        }

    def get_artifacts(self) -> ArtifactCollection:
        """Return model binaries"""
        artifacts = _RelativePathWrapper(self.model.artifact_any, MODEL_BIN_PATH)
        if len(self.server.additional_binaries) > 0:
            artifacts = CompositeArtifactCollection([
                artifacts,
                Blobs({os.path.basename(f): LocalFileBlob(f) for f in self.server.additional_binaries})
            ])
        return artifacts

    def get_python_version(self):
        """
        :return: version of python that produced model
        """
        version = self.model.params.get(Model.PYTHON_VERSION)
        return version or super(MLModelProvider, self).get_python_version()


class ModelBuildable(BuildableWithServer, WithMetadataRepository):
    def __init__(self, model_id: Union[int, Model], server_type: str, debug: bool = False):
        super().__init__(server_type)
        self.debug = debug
        if isinstance(model_id, int):
            self.model_id = model_id
            self.model_cache = None
        else:
            self.model_cache = model_id
            self.bind_meta_repo(self.model_cache._meta)
            self.model_id = model_id.id

    @property
    def task(self) -> Optional[Task]:
        return self.model.task

    @property
    def model(self):
        if self.model_cache is None:
            self._check_meta(False)
            self.model_cache = self._meta.get_model_by_id(self.model_id)
        return self.model_cache

    def get_provider(self) -> MLModelProvider:
        return MLModelProvider(self.model, self.server, self.debug)


class BuildableModelHook(BuildableHook, TypeHookMixin):
    valid_types = [Model]

    def process(self, obj, **kwargs):
        server = kwargs.get('server')  # TODO ???
        debug = kwargs.get('debug', False)
        return ModelBuildable(obj, server_type=server.type, debug=debug)



from typing import Dict

import sklearn
from pyjackson.decorators import make_string
from sklearn.base import ClassifierMixin, RegressorMixin

from algolink.core.analyzer.base import TypeHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects.requirements import InstallableRequirement, Requirements
from algolink.core.objects.wrapper import ModelWrapper, PickleModelIO
from algolink.utils.module import get_object_base_module


class SklearnModelWrapper(ModelWrapper):
    """
    `pickle`-based :class:`.ModelWrapper` implementation for `scikit-learn` models
    """
    def __init__(self):
        super().__init__(PickleModelIO())

    def _model_requirements(self) -> Requirements:
        if get_object_base_module(self.model) is sklearn:
            return Requirements([InstallableRequirement.from_module(sklearn)])
        # some sklearn compatible model (either from library or user code) - fallback
        return super()._model_requirements()

    def _exposed_methods_mapping(self) -> Dict[str, str]:
        ret = {
            'predict': 'predict'
        }
        if isinstance(self.model, ClassifierMixin):
            ret['predict_proba'] = 'predict_proba'
        return ret


@make_string(include_name=True)
class SklearnHook(BindingModelHook, TypeHookMixin):
    """
    :class:`algolink.core.analyzer.model.ModelHook` implementation for `scikit-learn` models
    which uses :class:`SklearnModelWrapper`
    """
    valid_types = [ClassifierMixin, RegressorMixin]

    def _wrapper_factory(self) -> ModelWrapper:
        return SklearnModelWrapper()



import io
from typing import Tuple

import numpy as np

from algolink.core.objects import ArtifactCollection
from algolink.core.objects.artifacts import LazyBlob
from algolink.core.objects.dataset_source import Dataset
from algolink.repository.dataset.artifact import DatasetReader, DatasetWriter

DATA_FILE = 'data.npz'
DATA_KEY = 'data'


def save_npz(data: np.array):
    buf = io.BytesIO()
    np.savez_compressed(buf, **{DATA_KEY: data})
    return buf


class NumpyNdarrayWriter(DatasetWriter):
    """DatasetWriter implementation for numpy ndarray"""

    def write(self, dataset: Dataset) -> Tuple[DatasetReader, ArtifactCollection]:
        return NumpyNdarrayReader(), ArtifactCollection.from_blobs(
            {DATA_FILE: LazyBlob(lambda: save_npz(dataset.data))})


class NumpyNdarrayReader(DatasetReader):
    """DatasetReader implementation for numpy ndarray"""

    def read(self, artifacts: ArtifactCollection) -> Dataset:
        with artifacts.blob_dict() as blobs:
            with blobs[DATA_FILE].bytestream() as f:
                data = np.load(f)[DATA_KEY]
        return Dataset.from_object(data)



import os
import re
import shutil
import sys
import tempfile

from jinja2 import Environment, FileSystemLoader

from algolink.ext.docker import DockerIORegistry
from algolink.ext.docker.utils import create_docker_client, image_exists_at_dockerhub, repository_tags_at_dockerhub
from algolink.utils.log import logger


def prebuild_image(prebuild_path, name_template, python_version, *, push=False):
    tag = name_template.format(python_version)
    if image_exists_at_dockerhub(tag):
        logger.info('Skipped building image %s: already exists', tag)
        return

    with tempfile.TemporaryDirectory() as tmpdir:
        context_dir = os.path.join(tmpdir, 'context')
        logger.info('Building image %s on %s ...', tag, context_dir)

        shutil.copytree(prebuild_path, context_dir)
        _generate_dockerfile(context_dir, python_version)

        try:
            with create_docker_client() as client:
                client.images.build(path=context_dir, tag=tag, rm=True)

            logger.info('Successfully built image %s', tag)
        except Exception as e:
            logger.error('Failed to build image %s: %s', tag, e)
            return

    if push:
        with create_docker_client() as client:
            DockerIORegistry().login(client)
            client.images.push(tag)


def _generate_dockerfile(context_dir, python_version):
    j2 = Environment(loader=FileSystemLoader([context_dir]))
    docker_tmpl = j2.get_template('Dockerfile.j2')
    with open(os.path.join(context_dir, 'Dockerfile'), 'w') as df:
        df.write(docker_tmpl.render({'python_version': python_version}))


PY_BLACK_LIST = {'^2\\.', '^3\\.[0-5]\\.', '^3\\.6\\.[0-4]'}


def prebuild_missing_images(prebuild_path, name_template):
    if len(sys.argv) < 2:
        python_tags = repository_tags_at_dockerhub('python')
        python_versions = [tag for tag in python_tags if re.match('^[0-9]+\\.[0-9]+\\.[0-9]+$', tag)]
        python_versions = [tag for tag in python_versions if not any(re.match(pre, tag) for pre in PY_BLACK_LIST)]
        python_versions.sort()
    else:
        python_versions = sys.argv[1:]

    for python_version in python_versions:
        prebuild_image(prebuild_path, name_template, python_version, push=True)



from typing import Any, Dict

from algolink.core.objects.core import Model


def create_model(model_object, input_data, model_name: str = None, params: Dict[str, Any] = None,
                 description: str = None, **kwargs) -> Model:
    """
    Creates Model instance from arbitrary model objects and sample of input data

    :param model_object: model object (function, sklearn model, tensorflow output tensor list etc)
    :param input_data: sample of input data (numpy array, pandas dataframe, feed dict etc)
    :param model_name: name for model in database, if not provided will be autogenerated
    :param params: dict with arbitrary parameters. Must be json-serializable
    :param description: text description of this model
    :param kwargs: other arguments for model (see Model.create)
    :return: :class:`~algolink.core.objects.core.Model` instance
    """
    return Model.create(model_object, input_data, model_name, params, description, **kwargs)



from abc import abstractmethod
from typing import Union

from pyjackson.core import ArgList, Field
from pyjackson.generics import Serializer


class TypeWithSpec(Serializer):
    """
    Abstract base class for types providing its OpenAPI schema definition
    """

    @abstractmethod
    def get_spec(self) -> ArgList:
        pass  # pragma: no cover

    def is_list(self):
        return False

    @abstractmethod
    def list_size(self):
        pass  # pragma: no cover


class ListTypeWithSpec(TypeWithSpec):
    """
    Abstract base class for `list`-like types providing its OpenAPI schema definition
    """

    def is_list(self):
        return True

    @abstractmethod
    def list_size(self):
        return NotImplementedError  # pragma: no cover


class SizedTypedListType(ListTypeWithSpec):
    """
    Subclass of :class:`ListTypeWithSpec` which specifies size of internal `list`
    """

    def __init__(self, size: Union[int, None], dtype: type):
        self.dtype = dtype
        self.size = size

    def get_spec(self) -> ArgList:
        return [Field(0, self.dtype, False)]

    def list_size(self):
        return self.size

    @abstractmethod
    def deserialize(self, obj):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def serialize(self, instance):
        raise NotImplementedError  # pragma: no cover



from abc import abstractmethod
from typing import List, Type

from .base import Hook, analyzer_class
from ..objects.wrapper import CallableMethodModelWrapper, ModelWrapper


class ModelHook(Hook):
    """
    Base hook type for :py:class:`ModelAnalyzer`.
    Analysis result is an instance of :class:`~algolink.core.objects.ModelWrapper`
    """
    valid_types: List[Type] = None

    @abstractmethod
    def process(self, obj, **kwargs) -> ModelWrapper:
        pass  # pragma: no cover


ModelAnalyzer = analyzer_class(ModelHook, ModelWrapper)


class BindingModelHook(ModelHook):
    """
    Binding model hook which `process` by first creating corresponding model wrapper (by means of a subclass) and
    then binding created wrapper to given model object
    """

    def process(self, obj, **kwargs) -> ModelWrapper:
        return self._wrapper_factory().bind_model(obj, **kwargs)

    @abstractmethod
    def _wrapper_factory(self) -> ModelWrapper:
        pass  # pragma: no cover


class CallableMethodModelHook(BindingModelHook):
    """
    Hook for processing functions
    """
    def _wrapper_factory(self) -> ModelWrapper:
        return CallableMethodModelWrapper()

    def can_process(self, obj) -> bool:
        return callable(obj)

    def must_process(self, obj) -> bool:
        return False



import contextlib
import os
import tempfile
import typing

import lightgbm as lgb

from algolink.core.analyzer.base import TypeHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects.artifacts import Blobs, LocalFileBlob
from algolink.core.objects.wrapper import FilesContextManager, LibModelWrapperMixin, ModelIO, ModelWrapper


class LightGBMModelIO(ModelIO):
    """
    :class:`.ModelIO` implementation for `lightgbm.Booster` type
    """
    model_path = 'model.lgb'

    @contextlib.contextmanager
    def dump(self, model: lgb.Booster) -> FilesContextManager:
        with tempfile.TemporaryDirectory(prefix='algolink_lightgbm_dump') as f:
            path = os.path.join(f, self.model_path)
            model.save_model(path)
            yield Blobs({self.model_path: LocalFileBlob(path)})

    def load(self, path):
        model_file = os.path.join(path, self.model_path)
        return lgb.Booster(model_file=model_file)


class LightGBMModelWrapper(LibModelWrapperMixin):
    """
    :class:`.ModelWrapper` implementation for `lightgbm.Booster` type
    """
    libraries = [lgb]

    def __init__(self):
        super().__init__(LightGBMModelIO())

    def _exposed_methods_mapping(self) -> typing.Dict[str, str]:
        return {
            'predict': '_predict'
        }

    @ModelWrapper.with_model
    def _predict(self, data):
        if isinstance(data, lgb.Dataset):
            data = data.data
        return self.model.predict(data)


class LightGBMModelHook(BindingModelHook, TypeHookMixin):
    """
    :class:`.ModelHook` implementation for `lightgbm.Booster` type
    """
    valid_types = [lgb.Booster]

    def _wrapper_factory(self) -> ModelWrapper:
        return LightGBMModelWrapper()



import contextlib
import json
import os
import tempfile
from abc import abstractmethod
from typing import Dict

import tensorflow as tf
from pyjackson.decorators import make_string

from algolink.core.analyzer.base import CanIsAMustHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects.artifacts import Blobs, InMemoryBlob, LocalFileBlob
from algolink.core.objects.wrapper import FilesContextManager, LibModelWrapperMixin, ModelIO, ModelWrapper

TF_MODEL_FILENAME = 'graph'


class _TfModel:
    def __init__(self, tensors, session=None):
        self.tensors = tensors
        self.__session = session

        if isinstance(tensors, list):
            self.tensor_names = [t.name for t in tensors]
        else:
            self.tensor_names = tensors.name

        self.is_frozen = self._is_graph_frozen()

    def predict(self, input_data):
        """
        Runs session and returns output tensor values

        :param input_data: data to predict
        :return: prediction
        """
        prediction = self.get_session().run(self.tensors, feed_dict=input_data)
        if isinstance(prediction, list):
            return {str(i): tensor for i, tensor in enumerate(prediction)}
        return prediction

    def get_session(self):
        session = self.__session or tf.get_default_session()
        if session is None:
            raise ValueError('Cant work with model without session, please use inside "with session.as_default()"')
        return session

    def close_session(self):
        if self.__session is not None:
            self.__session.close()

    @staticmethod
    def _is_graph_frozen() -> bool:
        """
        Checks if graph in current graph is frozen

        :return: `True` or `False`
        """
        from tensorflow.python.ops import variables
        return not bool(variables._all_saveable_objects())


class _TFDump:
    """
    Base class for tf dumping method
    """

    @abstractmethod
    def dump(self, session, path) -> FilesContextManager:
        pass  # pragma: no cover

    @abstractmethod
    def load(self, session, path) -> tf.Graph:
        pass  # pragma: no cover


class _Saver(_TFDump):
    """
    :class:`_TFDump` implementation with tf.train.Saver
    """

    @contextlib.contextmanager
    def dump(self, session, path) -> FilesContextManager:
        with session.as_default(), session.graph.as_default():
            saver = tf.train.Saver(save_relative_paths=True)
            saver.save(session, os.path.join(path, TF_MODEL_FILENAME))

        yield Blobs({
            name: LocalFileBlob(os.path.join(path, name)) for name in os.listdir(path)
        })

    def load(self, session, path):
        with session.as_default(), session.graph.as_default():
            saver = tf.train.import_meta_graph(os.path.join(path, TF_MODEL_FILENAME) + '.meta')
            saver.restore(session, os.path.join(path, TF_MODEL_FILENAME))


class _Protobuf(_TFDump):
    """
    :class:`_TFDump` implementation with tf.train.write_graph
    """

    @contextlib.contextmanager
    def dump(self, session, path) -> FilesContextManager:
        tf.train.write_graph(session.graph.as_graph_def(), path, TF_MODEL_FILENAME, as_text=False)
        yield Blobs({
            TF_MODEL_FILENAME: LocalFileBlob(os.path.join(path, TF_MODEL_FILENAME))
        })

    def load(self, session, path):
        with session.graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(os.path.join(path, TF_MODEL_FILENAME), 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')


class TFTensorModelIO(ModelIO):
    """
    :class:`algolink.core.objects.ModelIO` for tensorflow models. Model is a `_TfModel` instance
    """
    meta_json = 'meta.json'

    @contextlib.contextmanager
    def dump(self, model: _TfModel) -> FilesContextManager:
        """
        Dumps session to temporary directory and creates :class:`~algolink.core.objects.ArtifactCollection` from it

        :return: context manager with :class:`~algolink.core.objects.ArtifactCollection`
        """
        with tempfile.TemporaryDirectory(prefix='algolink_tensor_') as tempdir:
            dumper = self._get_dumper(model.is_frozen)
            with dumper.dump(model.get_session(), tempdir) as artifact:
                meta = json.dumps([model.tensor_names, model.is_frozen]).encode('utf-8')
                yield artifact + Blobs({self.meta_json: InMemoryBlob(meta)})

    def load(self, path) -> _TfModel:
        """
        Loads graph from path

        :param path: path to load from
        """
        with open(os.path.join(path, self.meta_json)) as f:
            tensor_names, is_frozen = json.load(f)
        dumper = self._get_dumper(is_frozen)

        graph = tf.Graph()
        session = tf.Session(graph=graph)

        dumper.load(session, path)

        if isinstance(tensor_names, list):
            tensors = [graph.get_tensor_by_name(n) for n in tensor_names]
        else:
            tensors = graph.get_tensor_by_name(tensor_names)

        return _TfModel(tensors, session)

    @staticmethod
    def _get_dumper(is_frozen):
        return _Protobuf() if is_frozen else _Saver()


class TFTensorModelWrapper(LibModelWrapperMixin):
    """
    :class:`algolink.core.objects.ModelWrapper` for tensorflow models. `.model` attribute is a list of output tensors
    """
    libraries = [tf]

    def __init__(self):
        super().__init__(TFTensorModelIO())

    def _exposed_methods_mapping(self) -> Dict[str, str]:
        return {
            'predict': 'predict'
        }

    def load(self, path):
        self._close_session_if_any()
        super().load(path)

    def bind_model(self, model, input_data=None, **kwargs):
        self._close_session_if_any()
        return super().bind_model(_TfModel(model), input_data, **kwargs)

    def _close_session_if_any(self):
        if self.model is not None:
            self.model.close_session()


@make_string(include_name=True)
class TFTensorHook(CanIsAMustHookMixin, BindingModelHook):
    """
    Hook for tensorflow models
    """
    valid_types = [tf.Tensor, list]

    # here we couldn't use `TypeHookMixin` as we expect lists of `tf.Tensor`s only
    def must_process(self, obj) -> bool:
        """
        Returns `True` if object is a tf.Tensor or list of tf.Tensors

        :param obj: obj to check
        :return: `True` or `False`
        """
        return isinstance(obj, tf.Tensor) or (isinstance(obj, list) and all(isinstance(o, tf.Tensor) for o in obj))

    def _wrapper_factory(self) -> ModelWrapper:
        """
        Creates :class:`TFTensorModelWrapper` for tensorflow model object

        :return: :class:`TFTensorModelWrapper` instance
        """
        return TFTensorModelWrapper()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import, print_function

import io
import os
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup

# allow setup.py to run from another directory
here = os.path.dirname(__file__)
here and os.chdir(here)


def read(*names, **kwargs):
    with io.open(
            join(dirname(__file__), *names),
            encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


def get_requirements(file_name):
    return [r for r in read(file_name).split('\n') if r and not r.startswith('#')]


setup_args = dict(
    name='algolink',
    version='0.7.9',
    license='Apache-2.0',
    description='Machine Learning Lifecycle Framework',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Leepand',
    author_email='leepand6@gmail.com',
    url='https://github.com/leepand/algolink',
    #packages=find_packages(exclude=["tests", "tests.*"]),#setuptools.find_packages(exclude=["tests*"]),#
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities',
    ],
    project_urls={
        'Changelog': 'https://github.com/leepand/algolink/blob/master/CHANGELOG.rst',
        'Issue Tracker': 'https://github.com/leepand/algolink/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.6',
    install_requires=get_requirements('requirements.txt'),
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
        'testing': get_requirements('test.requirements.txt')
    },
    setup_requires=[
        'pytest-runner',
    ],
    entry_points={
        'console_scripts': [
            'alink = algolink.cli:main',
        ]
    },
)

if __name__ == '__main__':
    setup(**setup_args)
